from dataclasses import dataclass
from enum import Enum
from typing import Dict

from opentrons_shared_data.util import StrEnum

from opentrons.drivers.command_builder import CommandBuilder


class GCODE(StrEnum):
    GET_RESET_REASON = "M114"
    GET_DEVICE_INFO = "M115"
    SET_SERIAL_NUMBER = "M996"
    ENTER_BOOTLOADER = "dfu"
    SET_LED = "M200"
    SET_PRESSURE_STATE = "M120"
    GET_PRESSURE_STATE = "M121"
    SET_PUMP_STATE = "M122"
    GET_PUMP_STATE = "M123"
    SET_VENT_STATE = "M124"

    def build_command(self) -> CommandBuilder:
        """Build command."""
        return CommandBuilder().add_gcode(self)


class HardwareRevision(Enum):
    """Hardware Revision."""

    NFF = "nff"


@dataclass
class VacuumModuleInfo:
    """Vacuum module info."""

    fw: str
    hw: HardwareRevision
    sn: str
    rr: int = 0

    def to_dict(self) -> Dict[str, str]:
        """Build vacuum module info."""
        return {
            "serial": self.sn,
            "version": self.fw,
            "model": self.hw.value,
            "reset_reason": str(self.rr),
        }


class LEDColor(Enum):
    """Vacuum Module LED Color."""

    WHITE = 0
    RED = 1
    GREEN = 2
    BLUE = 3
    YELLOW = 4

    @classmethod
    def from_name(cls, name: str) -> "LEDColor":
        match name.lower():
            case "red":
                return cls.RED
            case "green":
                return cls.GREEN
            case "blue":
                return cls.BLUE
            case "yellow":
                return cls.YELLOW
            case _:
                return cls.WHITE

    def to_name(self) -> "str":
        return self.name.lower()


class LEDPattern(Enum):
    """Vacuum Module LED Pattern."""

    STATIC = 0
    FLASH = 1
    PULSE = 2
    CONFIRM = 3


class VentState(Enum):
    """The State of the vent."""

    OPENED = 0
    CLOSED = 1


@dataclass
class PressureState:
    """Get the pressure state."""

    target_guage_pressure: float
    current_guage_pressure: float
    pressure_abs_a: float
    pressure_abs_b: float
    pressure_atm: float
    vacuum_enabled: bool
    vent_state: VentState


@dataclass
class PumpState:
    """Get the pump state."""

    target_rpm: float
    current_rpm: float
    target_pwm: float
    current_pwm: float
    pump_running: bool
    manual_control: bool



"""Tests for door_watcher."""

from typing import cast

import pytest
from anyio import to_thread
from decoy import Decoy, matchers

from opentrons.hardware_control import HardwareControlAPI, OT2HardwareControlAPI
from opentrons.hardware_control.types import (
    DoorState,
    DoorStateNotification,
    HardwareEventHandler,
    PauseType,
)
from opentrons.protocol_engine.actions import ActionDispatcher, DoorChangeAction
from opentrons.protocol_engine.execution.door_watcher import (
    DoorWatcher,
)
from opentrons.protocol_engine.state.state import StateStore


@pytest.fixture
def state_store(decoy: Decoy) -> StateStore:
    """Get a mock in the shape of a StateStore."""
    return decoy.mock(cls=StateStore)


@pytest.fixture
def hardware_control_api(
    decoy: Decoy,
) -> HardwareControlAPI:
    """Return a mock in the shape of a HardwareControlAPI."""
    return decoy.mock(cls=OT2HardwareControlAPI)


@pytest.fixture
def action_dispatcher(decoy: Decoy) -> ActionDispatcher:
    """Return a mock in the shape of an ActionDispatcher."""
    return decoy.mock(cls=ActionDispatcher)


@pytest.fixture
async def subject(
    state_store: StateStore,
    hardware_control_api: HardwareControlAPI,
    action_dispatcher: ActionDispatcher,
) -> DoorWatcher:
    """Return a DoorWatcher with mocked dependencies.

    Async because DoorWatcher's initializer requires a running event loop.
    """
    return DoorWatcher(
        state_store=state_store,
        hardware_api=hardware_control_api,
        action_dispatcher=action_dispatcher,
    )


async def test_event_forwarding(
    decoy: Decoy,
    subject: DoorWatcher,
    state_store: StateStore,
    hardware_control_api: HardwareControlAPI,
    action_dispatcher: ActionDispatcher,
) -> None:
    """It should forward events that come from a different thread."""
    handler_captor = matchers.Captor()
    unsubscribe_callback = decoy.mock(name="unsubscribe_callback")
    decoy.when(hardware_control_api.register_callback(handler_captor)).then_return(
        unsubscribe_callback
    )

    decoy.when(state_store.commands.get_is_running()).then_return(True)

    subject.start()

    captured_handler = cast(HardwareEventHandler, handler_captor.value)

    input_event = DoorStateNotification(new_state=DoorState.OPEN)
    expected_action_to_forward = DoorChangeAction(DoorState.OPEN)

    await to_thread.run_sync(captured_handler, input_event)
    decoy.verify(
        hardware_control_api.pause(PauseType.PAUSE),
        action_dispatcher.dispatch(expected_action_to_forward),
        times=1,
    )

    decoy.reset()
    input_event = DoorStateNotification(new_state=DoorState.CLOSED)
    await to_thread.run_sync(captured_handler, input_event)
    decoy.verify(
        hardware_control_api.pause(PauseType.PAUSE),
        times=0,
    )

    decoy.reset()

    decoy.when(state_store.commands.get_is_running()).then_return(True)

    input_event_module = DoorStateNotification(
        new_state=DoorState.OPEN, module_serial="magical_module"
    )
    expected_action_to_forward_module = DoorChangeAction(
        DoorState.OPEN, module_serial="magical_module"
    )

    await to_thread.run_sync(captured_handler, input_event_module)

    decoy.verify(
        hardware_control_api.pause(PauseType.PAUSE),
        action_dispatcher.dispatch(expected_action_to_forward_module),
        times=1,
    )

    decoy.reset()
    input_event_module = DoorStateNotification(
        new_state=DoorState.CLOSED, module_serial="magical_module"
    )
    await to_thread.run_sync(captured_handler, input_event_module)
    decoy.verify(
        hardware_control_api.pause(PauseType.PAUSE),
        times=0,
    )


async def test_one_subscribe_one_unsubscribe(
    decoy: Decoy,
    hardware_control_api: HardwareControlAPI,
    subject: DoorWatcher,
) -> None:
    """Multiple start()s and stop()s should be collapsed."""
    unsubscribe = decoy.mock(name="unsubscribe_callback")
    wrong_unsubscribe = decoy.mock(name="wrong_unsubscribe")

    decoy.when(hardware_control_api.register_callback(matchers.Anything())).then_return(
        unsubscribe, wrong_unsubscribe
    )

    subject.start()
    subject.start()
    subject.stop()
    subject.stop()

    decoy.verify(unsubscribe(), times=1)
    decoy.verify(wrong_unsubscribe(), times=0)



from opentrons import protocol_api, types

metadata = {
    "protocolName": "Testosaur Version 3",
    "author": "Opentrons <engineering@opentrons.com>",
    "description": 'A variant on "Dinosaur" for testing with Protocol API v3',
    "source": "Opentrons Repository",
    "apiLevel": "3.0",
}


def run(ctx: protocol_api.ProtocolContext) -> None:
    tip_rack = ctx.load_labware("opentrons_96_tiprack_300ul", 8)
    source = ctx.load_labware("nest_12_reservoir_15ml", 1)
    dest = ctx.load_labware("corning_96_wellplate_360ul_flat", 2)

    pipette = ctx.load_instrument("p300_single_gen2", types.Mount.RIGHT, [])

    for i in range(4):
        pipette.pick_up_tip(tip_rack.wells()[i])
        pipette.aspirate(50, source.wells_by_name()["A1"])
        pipette.dispense(50, dest.wells()[i])
        pipette.drop_tip(tip_rack.wells()[i])



import pytest

from opentrons.protocol_api.labware import Labware
from opentrons.types import DeckSlotName, Location, MeniscusTrackingTarget, Point


def test_point_from_xyz_attrs() -> None:
    class Original:
        def __init__(self) -> None:
            self.x = 1
            self.y = 2
            self.z = 3

    point = Point.from_xyz_attrs(Original())
    assert point == Point(1, 2, 3)


def test_point_mul() -> None:
    a = Point(1, 2, 3)
    b: float = 2
    assert a * b == Point(2, 4, 6)

    b = 3.1
    assert a * b == Point(3.1, 6.2, 9.3)

    with pytest.raises(TypeError):
        a * a  # type: ignore[operator]


def test_point_rmul() -> None:
    a = Point(1, 2, 3)
    b: float = 2
    assert b * a == Point(2, 4, 6)

    b = 3.1
    assert b * a == Point(3.1, 6.2, 9.3)


def test_location_repr_labware(min_lw: Labware) -> None:
    """It should represent labware as Labware"""
    loc = Location(point=Point(x=1.1, y=2.1, z=3.5), labware=min_lw)
    assert (
        f"{loc}"
        == "Location(point=Point(x=1.1, y=2.1, z=3.5), labware=minimal labware on deck, meniscus_tracking=None)"
    )


def test_location_repr_well(min_lw: Labware) -> None:
    """It should represent labware as Well"""
    loc = Location(point=Point(x=1, y=2, z=3), labware=min_lw.wells()[0])
    assert (
        f"{loc}"
        == "Location(point=Point(x=1, y=2, z=3), labware=A1 of minimal labware on deck, meniscus_tracking=None)"
    )


def test_location_repr_slot() -> None:
    """It should represent labware as a slot"""
    loc = Location(point=Point(x=-1, y=2, z=3), labware="1")
    assert (
        f"{loc}"
        == "Location(point=Point(x=-1, y=2, z=3), labware=1, meniscus_tracking=None)"
    )


@pytest.mark.parametrize(
    ("input", "expected_ot2_equivalent", "expected_ot3_equivalent"),
    [
        # Middle slot:
        (DeckSlotName.SLOT_5, DeckSlotName.SLOT_5, DeckSlotName.SLOT_C2),
        (DeckSlotName.SLOT_C2, DeckSlotName.SLOT_5, DeckSlotName.SLOT_C2),
        # Northwest corner:
        (DeckSlotName.SLOT_10, DeckSlotName.SLOT_10, DeckSlotName.SLOT_A1),
        (DeckSlotName.SLOT_A1, DeckSlotName.SLOT_10, DeckSlotName.SLOT_A1),
        # Northeast corner:
        (DeckSlotName.FIXED_TRASH, DeckSlotName.FIXED_TRASH, DeckSlotName.SLOT_A3),
        (DeckSlotName.SLOT_A3, DeckSlotName.FIXED_TRASH, DeckSlotName.SLOT_A3),
    ],
)
def test_deck_slot_name_equivalencies(
    input: DeckSlotName,
    expected_ot2_equivalent: DeckSlotName,
    expected_ot3_equivalent: DeckSlotName,
) -> None:
    assert (
        input.to_ot2_equivalent()
        == input.to_equivalent_for_robot_type("OT-2 Standard")
        == expected_ot2_equivalent
    )
    assert (
        input.to_ot3_equivalent()
        == input.to_equivalent_for_robot_type("OT-3 Standard")
        == expected_ot3_equivalent
    )


@pytest.mark.parametrize(
    ("input", "expected_int"),
    [
        (DeckSlotName.SLOT_5, 5),
        (DeckSlotName.SLOT_C2, 5),
    ],
)
def test_deck_slot_name_as_int(input: DeckSlotName, expected_int: int) -> None:
    assert input.as_int() == expected_int


def test_location_move_preseves_meniscus() -> None:
    """Location.move() should preserve meniscus relativeness."""
    loc = Location(
        point=Point(x=1, y=2, z=3),
        labware=None,
        _meniscus_tracking=MeniscusTrackingTarget.START,
    )
    moved = loc.move(Point(x=1, y=2, z=3))
    assert moved == Location(
        point=Point(x=2, y=4, z=6),
        labware=None,
        _meniscus_tracking=MeniscusTrackingTarget.START,
    )



from typing import Any, List, Tuple

import pytest
from _pytest.fixtures import SubRequest

from opentrons.hardware_control.modules.types import (
    MagneticModuleModel,
    TemperatureModuleModel,
    ThermocyclerModuleModel,
)
from opentrons.hardware_control.types import CriticalPoint
from opentrons.protocol_api import labware
from opentrons.protocol_api.core.legacy import module_geometry
from opentrons.protocol_api.core.legacy.deck import Deck
from opentrons.protocols.api_support.deck_type import (
    SHORT_TRASH_DECK,
    STANDARD_OT2_DECK,
)
from opentrons.protocols.geometry.planning import (
    plan_moves,
    safe_height,
    should_dodge_thermocycler,
)
from opentrons.types import Location, Point

tall_lw_name = "opentrons_96_tiprack_1000ul"
labware_name = "corning_96_wellplate_360ul_flat"
trough_name = "usascientific_12_reservoir_22ml"
P300M_GEN2_MAX_HEIGHT = 155.75


@pytest.fixture(
    # Limit the tests in this file to just test with OT-2 deck definitions.
    #
    # We need to do this because the tests in this file use the `Deck` class from
    # the older, non-Protocol-Engine versions of the Python Protocol API (apiLevel<=2.13),
    # and those versions do not support OT-3s.
    #
    # TODO(mm, 2023-05-18) We should either:
    #
    # * Decide that the functions tested here are only for PAPIv<=2.13.
    #   Then, we should move them to `opentrons.protocol_api.core.legacy` to indicate that.
    #
    # * Or, decide that the functions tested here are for all versions of PAPI.
    #   Then, we should rewrite these tests to not depend on the OT-2-only `Deck` class.
    params=[STANDARD_OT2_DECK, SHORT_TRASH_DECK]
)
def deck(request: SubRequest) -> Deck:
    return Deck(deck_type=request.param)


def check_arc_basic(
    arc: List[Tuple[Point, Any]], from_loc: Location, to_loc: Location
) -> None:
    """Check the tests that should always be true for different-well moves
    - we should always go only up, then only xy, then only down
    - we should have three moves
    """
    assert len(arc) == 3
    assert arc[0][0]._replace(z=0) == from_loc.point._replace(z=0)
    assert arc[0][0].z == arc[1][0].z
    assert arc[1][0]._replace(z=0) == to_loc.point._replace(z=0)
    assert arc[2][0] == to_loc.point
    assert arc[0][0].z >= from_loc.point.z
    assert arc[1][0].z >= to_loc.point.z


def test_direct_moves(deck: Deck) -> None:
    lw1 = labware.load(labware_name, deck.position_for(1))

    same_place = plan_moves(
        lw1.wells()[0].top(),
        lw1.wells()[0].top(),
        deck,
        instr_max_height=P300M_GEN2_MAX_HEIGHT,
    )
    assert same_place == [(lw1.wells()[0].top().point, None)]

    same_well = plan_moves(
        lw1.wells()[0].top(),
        lw1.wells()[0].bottom(),
        deck,
        instr_max_height=P300M_GEN2_MAX_HEIGHT,
    )
    assert same_well == [(lw1.wells()[0].bottom().point, None)]


def test_basic_arc(deck: Deck) -> None:
    lw1 = labware.load(labware_name, deck.position_for(1))
    lw2 = labware.load(labware_name, deck.position_for(2))
    deck[1] = lw1
    deck[2] = lw2

    # same-labware moves should use the smaller safe z
    same_lw = plan_moves(
        lw1.wells()[0].top(),
        lw1.wells()[8].bottom(),
        deck,
        P300M_GEN2_MAX_HEIGHT,
        5.0,
        10.0,
    )
    check_arc_basic(same_lw, lw1.wells()[0].top(), lw1.wells()[8].bottom())
    assert same_lw[0][0].z == lw1.wells()[0].top().point.z + 5.0

    # different-labware moves, or moves with no labware attached,
    # should use the larger safe z and the global z
    different_lw = plan_moves(
        lw1.wells()[0].top(),
        lw2.wells()[0].bottom(),
        deck,
        P300M_GEN2_MAX_HEIGHT,
        5.0,
        10.0,
    )
    check_arc_basic(different_lw, lw1.wells()[0].top(), lw2.wells()[0].bottom())
    assert different_lw[0][0].z == deck.highest_z + 10.0


def test_force_direct(deck: Deck) -> None:
    lw1 = labware.load(labware_name, deck.position_for(1))
    lw2 = labware.load(labware_name, deck.position_for(2))
    # same-labware moves should move direct
    same_lw = plan_moves(
        lw1.wells()[0].top(),
        lw1.wells()[8].bottom(),
        deck,
        P300M_GEN2_MAX_HEIGHT,
        5.0,
        10.0,
        force_direct=True,
    )
    assert same_lw == [(lw1.wells()[8].bottom().point, None)]

    # different-labware moves should move direct
    different_lw = plan_moves(
        lw1.wells()[0].top(),
        lw2.wells()[0].bottom(),
        deck,
        P300M_GEN2_MAX_HEIGHT,
        5.0,
        10.0,
        force_direct=True,
    )
    assert different_lw == [(lw2.wells()[0].bottom().point, None)]


def test_no_labware_loc(deck: Deck) -> None:
    labware_def = labware.get_labware_definition(labware_name)

    lw1 = labware.load(labware_name, deck.position_for(1))
    lw2 = labware.load(labware_name, deck.position_for(2))
    deck[1] = lw1
    deck[2] = lw2
    # Various flavors of locations without labware should work
    no_lw = Location(point=lw1.wells()[0].top().point, labware=None)

    no_from = plan_moves(
        no_lw, lw2.wells()[0].bottom(), deck, P300M_GEN2_MAX_HEIGHT, 5.0, 10.0
    )
    check_arc_basic(no_from, no_lw, lw2.wells()[0].bottom())
    assert no_from[0][0].z == deck.highest_z + 10.0

    no_to = plan_moves(
        lw1.wells()[0].bottom(), no_lw, deck, P300M_GEN2_MAX_HEIGHT, 5.0, 10.0
    )
    check_arc_basic(no_to, lw1.wells()[0].bottom(), no_lw)
    assert no_from[0][0].z == deck.highest_z + 10.0

    no_well = Location(point=lw1.wells()[0].top().point, labware=lw1)

    no_from_well = plan_moves(
        no_well, lw1.wells()[1].top(), deck, P300M_GEN2_MAX_HEIGHT, 5.0, 10.0
    )
    check_arc_basic(no_from_well, no_well, lw1.wells()[1].top())

    no_from_well_height = no_from_well[0][0].z
    assert labware_def["schemaVersion"] == 2  # For presence of ["dimensions"].
    lw_height_expected = labware_def["dimensions"]["zDimension"] + 5.0
    assert no_from_well_height == lw_height_expected

    no_to_well = plan_moves(
        lw1.wells()[1].top(), no_well, deck, P300M_GEN2_MAX_HEIGHT, 5.0, 10.0
    )
    check_arc_basic(no_to_well, lw1.wells()[1].top(), no_well)
    no_to_well_height = no_to_well[0][0].z
    assert no_to_well_height == lw_height_expected


def test_arc_tall_point(deck: Deck) -> None:
    lw1 = labware.load(labware_name, deck.position_for(1))
    tall_z = 100
    old_top = lw1.wells()[0].top()
    tall_point = old_top.point._replace(z=tall_z)
    tall_top = Location(point=tall_point, labware=old_top.labware)
    to_tall = plan_moves(lw1.wells()[2].top(), tall_top, deck, 5.0, 10.0)
    check_arc_basic(to_tall, lw1.wells()[2].top(), tall_top)
    assert to_tall[0][0].z == tall_z

    from_tall = plan_moves(tall_top, lw1.wells()[3].top(), deck, 5.0, 10.0)
    check_arc_basic(from_tall, tall_top, lw1.wells()[3].top())
    assert from_tall[0][0].z == tall_z

    no_well = Location(point=tall_top.point, labware=lw1)
    from_tall_lw = plan_moves(no_well, lw1.wells()[4].bottom(), deck, 5.0, 10.0)
    check_arc_basic(from_tall_lw, no_well, lw1.wells()[4].bottom())


def test_arc_lower_minimum_z_height(deck: Deck) -> None:
    lw1 = labware.load(labware_name, deck.position_for(1))
    tall_z = 100
    minimum_z_height = 42
    old_top = lw1.wells()[0].top()
    tall_point = old_top.point._replace(z=tall_z)
    tall_top = Location(point=tall_point, labware=old_top.labware)
    to_tall = plan_moves(
        lw1.wells()[2].top(),
        tall_top,
        deck,
        P300M_GEN2_MAX_HEIGHT,
        5.0,
        10.0,
        False,
        minimum_z_height=minimum_z_height,
    )
    check_arc_basic(to_tall, lw1.wells()[2].top(), tall_top)
    assert to_tall[0][0].z == tall_z

    from_tall = plan_moves(
        tall_top,
        lw1.wells()[3].top(),
        deck,
        P300M_GEN2_MAX_HEIGHT,
        5.0,
        10.0,
        minimum_z_height=minimum_z_height,
    )
    check_arc_basic(from_tall, tall_top, lw1.wells()[3].top())
    assert from_tall[0][0].z == tall_z

    no_well = Location(point=tall_top.point, labware=lw1)
    from_tall_lw = plan_moves(
        no_well, lw1.wells()[4].bottom(), deck, P300M_GEN2_MAX_HEIGHT, 5.0, 10.0
    )
    check_arc_basic(from_tall_lw, no_well, lw1.wells()[4].bottom())


def test_direct_minimum_z_height(deck: Deck) -> None:
    lw1 = labware.load(labware_name, deck.position_for(1))
    from_loc = lw1.wells()[0].bottom().move(Point(x=-2))
    to_loc = lw1.wells()[0].bottom().move(Point(x=2))
    zmo = 150
    # This would normally be a direct move since it’s inside the same well,
    # but we want to check that we override it into an arc
    moves = plan_moves(
        from_loc, to_loc, deck, P300M_GEN2_MAX_HEIGHT, minimum_z_height=zmo
    )
    assert len(moves) == 3
    assert moves[0][0].z == zmo  # equals zmo b/c 150 is max of all safe z's
    check_arc_basic(moves, from_loc, to_loc)


def test_direct_cp(deck: Deck) -> None:
    trough = labware.load(trough_name, deck.position_for(1))
    lw1 = labware.load(labware_name, deck.position_for(2))
    # when moving from no origin location to a centered labware we should
    # start in default cp
    from_nothing = plan_moves(
        Location(Point(50, 50, 50), None),
        trough.wells()[0].top(),
        deck,
        P300M_GEN2_MAX_HEIGHT,
    )
    check_arc_basic(
        from_nothing, Location(Point(50, 50, 50), None), trough.wells()[0].top()
    )
    assert from_nothing[0][1] is None
    assert from_nothing[1][1] == CriticalPoint.XY_CENTER
    assert from_nothing[2][1] == CriticalPoint.XY_CENTER
    # when moving from an origin with a centered labware to a dest with a
    # centered labware we should stay in centered the entire time, whether
    # arc
    from_centered_arc = plan_moves(
        trough.wells()[0].top(), trough.wells()[1].top(), deck, P300M_GEN2_MAX_HEIGHT
    )
    check_arc_basic(from_centered_arc, trough.wells()[0].top(), trough.wells()[1].top())
    assert from_centered_arc[0][1] == CriticalPoint.XY_CENTER
    assert from_centered_arc[1][1] == CriticalPoint.XY_CENTER
    assert from_centered_arc[2][1] == CriticalPoint.XY_CENTER
    # or direct
    from_centered_direct = plan_moves(
        trough.wells()[0].top(), trough.wells()[1].bottom(), deck, P300M_GEN2_MAX_HEIGHT
    )
    assert from_centered_direct[0][1] == CriticalPoint.XY_CENTER
    # when moving from centered to normal, only the first move should be
    # centered
    to_normal = plan_moves(
        trough.wells()[0].top(), lw1.wells()[0].top(), deck, P300M_GEN2_MAX_HEIGHT
    )
    check_arc_basic(to_normal, trough.wells()[0].top(), lw1.wells()[0].top())
    assert to_normal[0][1] == CriticalPoint.XY_CENTER
    assert to_normal[1][1] is None
    assert to_normal[2][1] is None


@pytest.mark.ot2_only  # Due to usage of create_geometry_for_ot2_deck().
def test_gen2_module_transforms(deck: Deck) -> None:
    tmod = module_geometry.create_geometry(
        definition=module_geometry.load_definition(
            TemperatureModuleModel.TEMPERATURE_V2
        ),
        parent=deck.position_for("1"),
        configuration=None,
    )
    assert tmod.labware_offset == Point(-1.45, -0.15, 80.09)
    tmod2 = module_geometry.create_geometry(
        definition=module_geometry.load_definition(
            TemperatureModuleModel.TEMPERATURE_V2
        ),
        parent=deck.position_for("3"),
        configuration=None,
    )
    assert tmod2.labware_offset == Point(1.15, -0.15, 80.09)

    mmod = module_geometry.create_geometry(
        definition=module_geometry.load_definition(MagneticModuleModel.MAGNETIC_V2),
        parent=deck.position_for("1"),
        configuration=None,
    )
    assert mmod.labware_offset == Point(-1.175, -0.125, 82.25)
    mmod2 = module_geometry.create_geometry(
        definition=module_geometry.load_definition(MagneticModuleModel.MAGNETIC_V2),
        parent=deck.position_for("3"),
        configuration=None,
    )
    assert mmod2.labware_offset == Point(1.425, -0.125, 82.25)


def test_instr_max_height(deck: Deck) -> None:
    fixed_trash = deck.get_fixed_trash()
    trough = labware.load(trough_name, deck.position_for(1))
    trough2 = labware.load(trough_name, deck.position_for(2))
    deck[1] = trough
    deck[2] = trough2

    # if the highest deck height is between 1 mm and 10 mm below
    # the max instrument achievable height, we use the max instrument
    # height as the safe height
    assert isinstance(fixed_trash, labware.Labware)
    instr_max_height = fixed_trash.wells()[0].top().point.z + 1
    height = safe_height(
        from_loc=trough.wells()[0].top(),
        to_loc=trough2.wells()[0].top(),
        deck=deck,
        instr_max_height=round(instr_max_height, 2),
        well_z_margin=5.0,
        lw_z_margin=10.0,
    )
    assert height == round(instr_max_height, 2)

    # if the highest deck height is > 10 mm below the max instrument
    # height, we use the lw_z_margin instead
    instr_max_height = fixed_trash.wells()[0].top().point.z + 30
    height2 = safe_height(
        from_loc=trough.wells()[0].top(),
        to_loc=trough2.wells()[0].top(),
        deck=deck,
        instr_max_height=round(instr_max_height, 2),
        well_z_margin=5.0,
        lw_z_margin=10.0,
    )
    assert height2 == round(fixed_trash.wells()[0].top().point.z, 2) + 10.0

    # it fails if the highest deck height is less than 1 mm below
    # the max instr achievable height
    instr_max_height = fixed_trash.wells()[0].top().point.z
    with pytest.raises(Exception):
        safe_height(
            from_loc=trough.wells()[0].top(),
            to_loc=trough2.wells()[0].top(),
            deck=deck,
            instr_max_height=round(instr_max_height, 2),
            well_z_margin=5.0,
            lw_z_margin=10.0,
        )


def test_should_dodge(deck: Deck) -> None:
    # with no tc loaded, doesn't matter what the positions are
    assert not should_dodge_thermocycler(
        deck, deck.position_for(4), deck.position_for(9)
    )
    deck[7] = module_geometry.create_geometry(
        definition=module_geometry.load_definition(
            ThermocyclerModuleModel.THERMOCYCLER_V1
        ),
        parent=deck.position_for(7),
        configuration=None,
    )
    # with a tc loaded, some positions should require dodging
    assert should_dodge_thermocycler(deck, deck.position_for(12), deck.position_for(1))
    assert should_dodge_thermocycler(deck, deck.position_for(9), deck.position_for(4))
    # but ones that weren't explicitly marked shouldn't
    assert not should_dodge_thermocycler(
        deck, deck.position_for(1), deck.position_for(2)
    )
    # including a situation where we might have some messed up locations
    # with no parent
    assert not should_dodge_thermocycler(
        deck,
        Location(point=deck.position_for(1).point, labware=None),
        deck.position_for(12),
    )


def test_labware_in_next_slot(deck: Deck) -> None:
    trough = labware.load(trough_name, deck.position_for(4))
    trough2 = labware.load(trough_name, deck.position_for(1))
    trough3 = labware.load(trough_name, deck.position_for(3))
    deck[4] = trough
    deck[1] = trough2
    deck[3] = trough3
    assert deck.right_of("3") is None
    assert deck.left_of("2") is trough2
    assert deck.right_of("2") is trough3

    assert deck.right_of("9") is None


def test_get_non_fixture_slots(deck: Deck) -> None:
    trough = labware.load(trough_name, deck.position_for(4))
    deck[4] = trough

    assert deck.get_non_fixture_slots() == [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11]


def test_thermocycler_present(deck: Deck) -> None:
    """It should change when thermocycler is added/removed"""
    # Empty deck. No thermocycler
    assert not deck.thermocycler_present

    # Add a thermocycler
    deck[7] = module_geometry.create_geometry(
        definition=module_geometry.load_definition(
            ThermocyclerModuleModel.THERMOCYCLER_V1
        ),
        parent=deck.position_for(7),
        configuration=None,
    )
    assert deck.thermocycler_present

    # Add another labware
    deck[4] = labware.load(trough_name, deck.position_for(4))
    assert deck.thermocycler_present

    # Remove thermocycler
    del deck[7]
    assert not deck.thermocycler_present



"""Test Thermocycler wait for lid temperature command implementation."""

from decoy import Decoy

from opentrons.hardware_control.modules import Thermocycler
from opentrons.protocol_engine.commands import thermocycler as tc_commands
from opentrons.protocol_engine.commands.command import SuccessData
from opentrons.protocol_engine.commands.thermocycler.wait_for_lid_temperature import (
    WaitForLidTemperatureImpl,
)
from opentrons.protocol_engine.execution import EquipmentHandler
from opentrons.protocol_engine.state.module_substates import (
    ThermocyclerModuleId,
    ThermocyclerModuleSubState,
)
from opentrons.protocol_engine.state.state import StateView


async def test_set_target_block_temperature(
    decoy: Decoy,
    state_view: StateView,
    equipment: EquipmentHandler,
) -> None:
    """It should be able to wait for the specified module's target temperature."""
    subject = WaitForLidTemperatureImpl(state_view=state_view, equipment=equipment)

    data = tc_commands.WaitForLidTemperatureParams(
        moduleId="input-thermocycler-id",
    )
    expected_result = tc_commands.WaitForLidTemperatureResult()

    tc_module_substate = decoy.mock(cls=ThermocyclerModuleSubState)
    tc_hardware = decoy.mock(cls=Thermocycler)

    decoy.when(
        state_view.modules.get_thermocycler_module_substate("input-thermocycler-id")
    ).then_return(tc_module_substate)

    decoy.when(tc_module_substate.module_id).then_return(
        ThermocyclerModuleId("thermocycler-id")
    )

    # Get attached hardware modules
    decoy.when(
        equipment.get_module_hardware_api(ThermocyclerModuleId("thermocycler-id"))
    ).then_return(tc_hardware)

    result = await subject.execute(data)

    decoy.verify(
        tc_module_substate.get_target_lid_temperature(),
        await tc_hardware.wait_for_lid_target(),
    )
    assert result == SuccessData(public=expected_result)



"""Smoke tests for the AbstractRunner and ProtocolEngine classes.

These tests construct a AbstractRunner with a real ProtocolEngine
hooked to a simulating HardwareAPI.

Minimal, but valid and complete, protocol files are then loaded from
disk into the runner, and the protocols are run to completion. From
there, the ProtocolEngine state is inspected to check that
everything was loaded and run as expected.
"""

from datetime import datetime
from pathlib import Path

from decoy import matchers

from opentrons_shared_data.pipette.types import PipetteNameType

from opentrons.protocol_engine import (
    DeckPoint,
    DeckSlotLocation,
    EngineStatus,
    LoadedLabware,
    LoadedModule,
    LoadedPipette,
    ModuleDefinition,
    ModuleModel,
    commands,
)
from opentrons.protocol_reader import ProtocolReader
from opentrons.protocol_runner.create_simulating_orchestrator import (
    create_simulating_orchestrator,
)
from opentrons.types import DeckSlotName, MountType


async def test_runner_with_python(
    python_protocol_file: Path,
    tempdeck_v1_def: ModuleDefinition,
) -> None:
    """It should run a Python protocol on the PythonAndLegacyRunner."""
    protocol_reader = ProtocolReader()
    protocol_source = await protocol_reader.read_saved(
        files=[python_protocol_file],
        directory=None,
    )

    subject = await create_simulating_orchestrator(
        robot_type="OT-2 Standard", protocol_config=protocol_source.config
    )
    result = await subject.run(
        deck_configuration=[],
        protocol_source=protocol_source,
        run_time_param_values=None,
    )
    commands_result = result.commands
    pipettes_result = result.state_summary.pipettes
    labware_result = result.state_summary.labware
    modules_result = result.state_summary.modules

    pipette_id_captor = matchers.Captor()
    labware_id_captor = matchers.Captor()

    expected_pipette = LoadedPipette.model_construct(
        id=pipette_id_captor,
        pipetteName=PipetteNameType.P300_SINGLE,
        mount=MountType.LEFT,
    )

    expected_labware = LoadedLabware.model_construct(
        id=labware_id_captor,
        location=DeckSlotLocation(slotName=DeckSlotName.SLOT_1),
        loadName="opentrons_96_tiprack_300ul",
        definitionUri="opentrons/opentrons_96_tiprack_300ul/1",
        # fixme(mm, 2021-11-11): We should smoke-test that the engine picks up labware
        # offsets, but it's unclear to me what the best way of doing that is, since
        # we don't have access to the engine here to add offsets to it.
        offsetId=None,
    )

    expected_module = LoadedModule.model_construct(
        id=matchers.IsA(str),
        model=ModuleModel.TEMPERATURE_MODULE_V1,
        location=DeckSlotLocation(slotName=DeckSlotName.SLOT_3),
        serialNumber=matchers.IsA(str),
    )

    assert expected_pipette in pipettes_result
    assert expected_labware in labware_result
    assert expected_module in modules_result

    expected_command = commands.PickUpTip.model_construct(
        id=matchers.IsA(str),
        key=matchers.IsA(str),
        status=commands.CommandStatus.SUCCEEDED,
        createdAt=matchers.IsA(datetime),
        startedAt=matchers.IsA(datetime),
        completedAt=matchers.IsA(datetime),
        params=commands.PickUpTipParams(
            pipetteId=pipette_id_captor.value,
            labwareId=labware_id_captor.value,
            wellName="A1",
        ),
        notes=[],
        result=commands.PickUpTipResult(
            tipVolume=300.0,
            tipLength=51.83,
            tipDiameter=5.23,
            position=DeckPoint(x=14.38, y=74.24, z=64.69),
        ),
    )

    assert expected_command in commands_result
    await subject.finish()


async def test_runner_with_json(json_protocol_file: Path) -> None:
    """It should run a JSON protocol on the JsonRunner."""
    protocol_reader = ProtocolReader()
    protocol_source = await protocol_reader.read_saved(
        files=[json_protocol_file],
        directory=None,
    )

    subject = await create_simulating_orchestrator(
        robot_type="OT-2 Standard", protocol_config=protocol_source.config
    )
    result = await subject.run(deck_configuration=[], protocol_source=protocol_source)

    commands_result = result.commands
    pipettes_result = result.state_summary.pipettes
    labware_result = result.state_summary.labware

    expected_pipette = LoadedPipette(
        id="pipette-id",
        pipetteName=PipetteNameType.P300_SINGLE,
        mount=MountType.LEFT,
    )

    expected_labware = LoadedLabware(
        id="labware-id",
        location=DeckSlotLocation(slotName=DeckSlotName.SLOT_1),
        loadName="opentrons_96_tiprack_300ul",
        definitionUri="opentrons/opentrons_96_tiprack_300ul/1",
        displayName="Opentrons 96 Tip Rack 300 µL",
        # fixme(mm, 2021-11-11): We should smoke-test that the engine picks up labware
        # offsets, but it's unclear to me what the best way of doing that is, since
        # we don't have access to the engine here to add offsets to it.
        offsetId=None,
    )

    assert expected_pipette in pipettes_result
    assert expected_labware in labware_result

    expected_command = commands.PickUpTip.model_construct(
        id=matchers.IsA(str),
        key=matchers.IsA(str),
        status=commands.CommandStatus.SUCCEEDED,
        createdAt=matchers.IsA(datetime),
        startedAt=matchers.IsA(datetime),
        completedAt=matchers.IsA(datetime),
        params=commands.PickUpTipParams(
            pipetteId="pipette-id",
            labwareId="labware-id",
            wellName="A1",
        ),
        notes=[],
        result=commands.PickUpTipResult(
            tipVolume=300.0,
            tipLength=51.83,
            tipDiameter=5.23,
            position=DeckPoint(x=14.38, y=74.24, z=64.69),
        ),
    )

    assert expected_command in commands_result
    await subject.finish()


async def test_runner_with_legacy_python(legacy_python_protocol_file: Path) -> None:
    """It should run a Python protocol on the PythonAndLegacyRunner."""
    protocol_reader = ProtocolReader()
    protocol_source = await protocol_reader.read_saved(
        files=[legacy_python_protocol_file],
        directory=None,
    )

    subject = await create_simulating_orchestrator(
        robot_type="OT-2 Standard", protocol_config=protocol_source.config
    )
    result = await subject.run(
        deck_configuration=[],
        protocol_source=protocol_source,
        run_time_param_values=None,
    )

    commands_result = result.commands
    pipettes_result = result.state_summary.pipettes
    labware_result = result.state_summary.labware

    pipette_id_captor = matchers.Captor()
    labware_id_captor = matchers.Captor()

    expected_pipette = LoadedPipette.model_construct(
        id=pipette_id_captor,
        pipetteName=PipetteNameType.P300_SINGLE,
        mount=MountType.LEFT,
    )

    expected_labware = LoadedLabware.model_construct(
        id=labware_id_captor,
        location=DeckSlotLocation(slotName=DeckSlotName.SLOT_1),
        loadName="opentrons_96_tiprack_300ul",
        definitionUri="opentrons/opentrons_96_tiprack_300ul/1",
        # fixme(mm, 2021-11-11): When legacy running supports labware offsets, check
        # for that here.
        offsetId=None,
    )

    assert expected_pipette in pipettes_result
    assert expected_labware in labware_result

    expected_command = commands.PickUpTip.model_construct(
        id=matchers.IsA(str),
        key=matchers.IsA(str),
        status=commands.CommandStatus.SUCCEEDED,
        createdAt=matchers.IsA(datetime),
        startedAt=matchers.IsA(datetime),
        completedAt=matchers.IsA(datetime),
        params=commands.PickUpTipParams(
            pipetteId=pipette_id_captor.value,
            labwareId=labware_id_captor.value,
            wellName="A1",
        ),
        notes=[],
        result=commands.PickUpTipResult(
            tipVolume=300.0, tipLength=51.83, position=DeckPoint(x=0, y=0, z=0)
        ),
    )

    assert expected_command in commands_result
    await subject.finish()


async def test_runner_with_legacy_json(legacy_json_protocol_file: Path) -> None:
    """It should run a Python protocol on the PythonAndLegacyRunner."""
    protocol_reader = ProtocolReader()
    protocol_source = await protocol_reader.read_saved(
        files=[legacy_json_protocol_file],
        directory=None,
    )

    subject = await create_simulating_orchestrator(
        robot_type="OT-2 Standard", protocol_config=protocol_source.config
    )
    result = await subject.run(
        deck_configuration=[],
        protocol_source=protocol_source,
        run_time_param_values=None,
    )

    commands_result = result.commands
    pipettes_result = result.state_summary.pipettes
    labware_result = result.state_summary.labware

    pipette_id_captor = matchers.Captor()
    labware_id_captor = matchers.Captor()

    expected_pipette = LoadedPipette.model_construct(
        id=pipette_id_captor,
        pipetteName=PipetteNameType.P300_SINGLE,
        mount=MountType.LEFT,
    )

    expected_labware = LoadedLabware.model_construct(
        id=labware_id_captor,
        location=DeckSlotLocation(slotName=DeckSlotName.SLOT_1),
        loadName="opentrons_96_tiprack_300ul",
        definitionUri="opentrons/opentrons_96_tiprack_300ul/1",
        displayName="Opentrons 96 Tip Rack 300 µL",
        # fixme(mm, 2021-11-11): When legacy running supports labware offsets, check
        # for that here.
        offsetId=None,
    )

    assert expected_pipette in pipettes_result
    assert expected_labware in labware_result

    expected_command = commands.PickUpTip.model_construct(
        id=matchers.IsA(str),
        key=matchers.IsA(str),
        status=commands.CommandStatus.SUCCEEDED,
        createdAt=matchers.IsA(datetime),
        startedAt=matchers.IsA(datetime),
        completedAt=matchers.IsA(datetime),
        params=commands.PickUpTipParams(
            pipetteId=pipette_id_captor.value,
            labwareId=labware_id_captor.value,
            wellName="A1",
        ),
        notes=[],
        result=commands.PickUpTipResult(
            tipVolume=300.0, tipLength=51.83, position=DeckPoint(x=0, y=0, z=0)
        ),
    )

    assert expected_command in commands_result
    await subject.finish()


async def test_runner_with_python_and_run_time_parameters(
    python_protocol_file_with_run_time_params: Path,
) -> None:
    """It should run a Python protocol on the PythonAndLegacyRunner."""
    protocol_reader = ProtocolReader()
    protocol_source = await protocol_reader.read_saved(
        files=[python_protocol_file_with_run_time_params],
        directory=None,
    )

    subject = await create_simulating_orchestrator(
        robot_type="OT-2 Standard", protocol_config=protocol_source.config
    )
    result = await subject.run(
        deck_configuration=[],
        protocol_source=protocol_source,
        run_time_param_values={"aspirate_volume": 40.2},
    )
    commands_result = result.commands
    pipettes_result = result.state_summary.pipettes
    tiprack_result = result.state_summary.labware

    pipette_id_captor = matchers.Captor()
    tiprack_id_captor = matchers.Captor()
    reservoir_id_captor = matchers.Captor()

    expected_pipette = LoadedPipette.model_construct(
        id=pipette_id_captor,
        pipetteName=PipetteNameType.P300_SINGLE,
        mount=MountType.LEFT,
    )

    expected_tiprack = LoadedLabware.model_construct(
        id=tiprack_id_captor,
        location=DeckSlotLocation(slotName=DeckSlotName.SLOT_1),
        loadName="opentrons_96_tiprack_300ul",
        definitionUri="opentrons/opentrons_96_tiprack_300ul/1",
        # fixme(mm, 2021-11-11): We should smoke-test that the engine picks up labware
        # offsets, but it's unclear to me what the best way of doing that is, since
        # we don't have access to the engine here to add offsets to it.
        offsetId=None,
    )

    expected_reservoir = LoadedLabware.model_construct(
        id=reservoir_id_captor,
        location=DeckSlotLocation(slotName=DeckSlotName.SLOT_2),
        loadName="nest_1_reservoir_195ml",
        definitionUri="opentrons/nest_1_reservoir_195ml/2",
        # fixme(mm, 2021-11-11): We should smoke-test that the engine picks up labware
        # offsets, but it's unclear to me what the best way of doing that is, since
        # we don't have access to the engine here to add offsets to it.
        offsetId=None,
    )

    assert expected_pipette in pipettes_result
    assert expected_tiprack in tiprack_result
    assert expected_reservoir in tiprack_result

    assert result.state_summary.status == EngineStatus.SUCCEEDED

    expected_command = commands.Aspirate.model_construct(
        id=matchers.IsA(str),
        key=matchers.IsA(str),
        status=commands.CommandStatus.SUCCEEDED,
        createdAt=matchers.IsA(datetime),
        startedAt=matchers.IsA(datetime),
        completedAt=matchers.IsA(datetime),
        params=commands.AspirateParams.model_construct(
            labwareId=reservoir_id_captor.value,
            wellName=matchers.IsA(str),
            wellLocation=matchers.Anything(),
            flowRate=matchers.IsA(float),
            volume=40.2,
            pipetteId=pipette_id_captor.value,
        ),
        notes=[],
        result=matchers.Anything(),
    )

    assert expected_command in commands_result
    await subject.finish()



"""Tests for thermocycler_movement_flagger."""

from contextlib import nullcontext as does_not_raise
from typing import Any, ContextManager, NamedTuple, Optional

import pytest
from decoy import Decoy

from opentrons.drivers.types import ThermocyclerLidStatus
from opentrons.hardware_control import API as HardwareAPI
from opentrons.hardware_control.modules import Thermocycler as HardwareThermocycler
from opentrons.protocol_engine.errors import (
    ThermocyclerNotOpenError,
    WrongModuleTypeError,
)
from opentrons.protocol_engine.execution.equipment import EquipmentHandler
from opentrons.protocol_engine.execution.thermocycler_movement_flagger import (
    ThermocyclerMovementFlagger,
)
from opentrons.protocol_engine.state.module_substates.thermocycler_module_substate import (
    ThermocyclerModuleId,
    ThermocyclerModuleSubState,
)
from opentrons.protocol_engine.state.state import StateStore
from opentrons.protocol_engine.types import (
    DeckSlotLocation,
    ModuleLocation,
)
from opentrons.types import DeckSlotName


@pytest.fixture
def hardware_api(decoy: Decoy) -> HardwareAPI:
    """Get a mock in the shape of a HardwareAPI."""
    return decoy.mock(cls=HardwareAPI)


@pytest.fixture
def state_store(decoy: Decoy) -> StateStore:
    """Get a mock in the shape of a StateStore."""
    return decoy.mock(cls=StateStore)


@pytest.fixture
def equipment_handler(decoy: Decoy) -> EquipmentHandler:
    """Get a mock in the shape of an EquipmentHandler."""
    return decoy.mock(cls=EquipmentHandler)


@pytest.fixture
def subject(
    state_store: StateStore,
    hardware_api: HardwareAPI,
    equipment_handler: EquipmentHandler,
) -> ThermocyclerMovementFlagger:
    """Return a movement flagger initialized with mocked-out dependencies."""
    return ThermocyclerMovementFlagger(
        state_store=state_store, hardware_api=hardware_api, equipment=equipment_handler
    )


async def test_raises_depending_on_thermocycler_substate_lid_status(
    subject: ThermocyclerMovementFlagger,
    state_store: StateStore,
    decoy: Decoy,
) -> None:
    """It should flag movement depending on engine's thermocycler state."""
    decoy.when(
        state_store.modules.get_thermocycler_module_substate(module_id="module-id")
    ).then_return(
        ThermocyclerModuleSubState(
            module_id=ThermocyclerModuleId("module-id"),
            is_lid_open=False,
            target_lid_temperature=None,
            target_block_temperature=None,
        ),
    )

    with pytest.raises(ThermocyclerNotOpenError):
        await subject.ensure_labware_in_open_thermocycler(
            labware_parent=ModuleLocation(moduleId="module-id"),
        )


class LidStatusAndRaiseSpec(NamedTuple):
    """Test parametrization data.

    A Thermocycler lid status, and what we expect the subject to raise when it finds
    that the Thermocycler has that lid status.
    """

    # Optional to match current signature of Thermocycler.lid_status.
    # Should change to non-Optional if/when that becomes non-optional.
    lid_status: Optional[ThermocyclerLidStatus]
    expected_raise_cm: ContextManager[Any]


@pytest.mark.parametrize(
    LidStatusAndRaiseSpec._fields,
    [
        LidStatusAndRaiseSpec(
            lid_status=ThermocyclerLidStatus.CLOSED,
            expected_raise_cm=pytest.raises(ThermocyclerNotOpenError),
        ),
        LidStatusAndRaiseSpec(
            lid_status=ThermocyclerLidStatus.IN_BETWEEN,
            expected_raise_cm=pytest.raises(ThermocyclerNotOpenError),
        ),
        LidStatusAndRaiseSpec(
            lid_status=ThermocyclerLidStatus.UNKNOWN,
            expected_raise_cm=pytest.raises(ThermocyclerNotOpenError),
        ),
        LidStatusAndRaiseSpec(
            lid_status=ThermocyclerLidStatus.OPEN,
            expected_raise_cm=does_not_raise(),
        ),
    ],
)
async def test_raises_depending_on_thermocycler_hardware_lid_status(
    lid_status: ThermocyclerLidStatus,
    expected_raise_cm: ContextManager[Any],
    subject: ThermocyclerMovementFlagger,
    state_store: StateStore,
    hardware_api: HardwareAPI,
    decoy: Decoy,
) -> None:
    """When on a Thermocycler, it should raise if the lid isn't open."""
    decoy.when(
        state_store.modules.get_thermocycler_module_substate(module_id="module-id")
    ).then_return(
        ThermocyclerModuleSubState(
            module_id=ThermocyclerModuleId("module-id"),
            is_lid_open=True,
            target_lid_temperature=None,
            target_block_temperature=None,
        ),
    )

    decoy.when(state_store.config.use_virtual_modules).then_return(False)
    decoy.when(
        state_store.modules.get_serial_number(module_id="module-id")
    ).then_return("module-serial")

    thermocycler = decoy.mock(cls=HardwareThermocycler)
    decoy.when(thermocycler.device_info).then_return({"serial": "module-serial"})
    decoy.when(thermocycler.lid_status).then_return(lid_status)
    decoy.when(hardware_api.attached_modules).then_return([thermocycler])
    decoy.when(
        subject._equipment.get_module_hardware_api(ThermocyclerModuleId("module-id"))
    ).then_return(thermocycler)
    with expected_raise_cm:
        await subject.ensure_labware_in_open_thermocycler(
            labware_parent=ModuleLocation(moduleId="module-id"),
        )


async def test_raises_if_hardware_module_has_gone_missing(
    subject: ThermocyclerMovementFlagger,
    state_store: StateStore,
    hardware_api: HardwareAPI,
    decoy: Decoy,
) -> None:
    """It should raise if the hardware module can't be found by its serial no."""
    decoy.when(
        state_store.modules.get_thermocycler_module_substate(module_id="module-id")
    ).then_return(
        ThermocyclerModuleSubState(
            module_id=ThermocyclerModuleId("module-id"),
            is_lid_open=True,
            target_lid_temperature=None,
            target_block_temperature=None,
        ),
    )

    decoy.when(state_store.config.use_virtual_modules).then_return(False)
    decoy.when(
        state_store.modules.get_serial_number(module_id="module-id")
    ).then_return("module-serial")
    decoy.when(hardware_api.attached_modules).then_return([])

    with pytest.raises(ThermocyclerNotOpenError):
        await subject.ensure_labware_in_open_thermocycler(
            labware_parent=ModuleLocation(moduleId="module-id"),
        )


async def test_passes_if_virtual_module_lid_open(
    subject: ThermocyclerMovementFlagger,
    state_store: StateStore,
    decoy: Decoy,
) -> None:
    """It shouldn't raise if a virtual module in analysis has lid open."""
    decoy.when(
        state_store.modules.get_thermocycler_module_substate(module_id="module-id")
    ).then_return(
        ThermocyclerModuleSubState(
            module_id=ThermocyclerModuleId("module-id"),
            is_lid_open=True,
            target_lid_temperature=None,
            target_block_temperature=None,
        ),
    )
    decoy.when(state_store.config.use_virtual_modules).then_return(True)
    await subject.ensure_labware_in_open_thermocycler(
        labware_parent=ModuleLocation(moduleId="module-id")
    )


async def test_passes_if_labware_on_non_thermocycler_module(
    subject: ThermocyclerMovementFlagger,
    state_store: StateStore,
    hardware_api: HardwareAPI,
    decoy: Decoy,
) -> None:
    """It shouldn't raise if the labware is on a module other than a Thermocycler."""
    decoy.when(
        state_store.modules.get_thermocycler_module_substate(module_id="module-id")
    ).then_raise(WrongModuleTypeError("Woops"))
    await subject.ensure_labware_in_open_thermocycler(
        ModuleLocation(moduleId="module-id")
    )


async def test_passes_if_labware_not_on_any_module(
    subject: ThermocyclerMovementFlagger,
    state_store: StateStore,
    hardware_api: HardwareAPI,
    decoy: Decoy,
) -> None:
    """It shouldn't raise if the labware isn't on a module."""
    await subject.ensure_labware_in_open_thermocycler(
        DeckSlotLocation(slotName=DeckSlotName.SLOT_1)
    )



"""Test Flex Stacker store command implementation."""

from datetime import datetime
from typing import Type, Union
from unittest.mock import sentinel

import pytest
from decoy import Decoy, matchers

from opentrons_shared_data.errors.exceptions import (
    FlexStackerShuttleLabwareError,
    FlexStackerShuttleMissingError,
    FlexStackerStallError,
)
from opentrons_shared_data.labware.labware_definition import LabwareDefinition

from opentrons.drivers.flex_stacker.types import StackerAxis
from opentrons.hardware_control.modules import FlexStacker, PlatformState
from opentrons.protocol_engine.commands import flex_stacker
from opentrons.protocol_engine.commands.command import DefinedErrorData, SuccessData
from opentrons.protocol_engine.commands.flex_stacker.common import (
    FlexStackerLabwareStoreError,
    FlexStackerShuttleError,
    FlexStackerStallOrCollisionError,
)
from opentrons.protocol_engine.commands.flex_stacker.store import StoreImpl
from opentrons.protocol_engine.errors import (
    CannotPerformModuleAction,
    FlexStackerLabwarePoolNotYetDefinedError,
    LabwareNotLoadedOnModuleError,
)
from opentrons.protocol_engine.execution import EquipmentHandler
from opentrons.protocol_engine.resources import ModelUtils
from opentrons.protocol_engine.state.module_substates import (
    FlexStackerId,
    FlexStackerSubState,
)
from opentrons.protocol_engine.state.state import StateView
from opentrons.protocol_engine.state.update_types import (
    BatchLabwareLocationUpdate,
    FlexStackerStateUpdate,
    StateUpdate,
)
from opentrons.protocol_engine.types import (
    InStackerHopperLocation,
    ModuleLocation,
    OnAddressableAreaLocationSequenceComponent,
    OnCutoutFixtureLocationSequenceComponent,
    OnModuleLocationSequenceComponent,
    StackerLabwareMovementStrategy,
    StackerStoredLabwareGroup,
)


@pytest.fixture
def subject(
    equipment: EquipmentHandler,
    state_view: StateView,
    model_utils: ModelUtils,
) -> StoreImpl:
    """Subject under test."""
    return StoreImpl(
        state_view=state_view, equipment=equipment, model_utils=model_utils
    )


def _contained_labware(count: int) -> list[StackerStoredLabwareGroup]:
    return [
        StackerStoredLabwareGroup(
            primaryLabwareId=f"primary-id-{i + 1}",
            adapterLabwareId=None,
            lidLabwareId=None,
        )
        for i in range(count)
    ]


async def test_store_raises_if_full(
    decoy: Decoy,
    equipment: EquipmentHandler,
    state_view: StateView,
    subject: StoreImpl,
    stacker_id: FlexStackerId,
    flex_50uL_tiprack: LabwareDefinition,
) -> None:
    """It should raise if called when the stacker is full."""
    data = flex_stacker.StoreParams(
        moduleId=stacker_id, strategy=StackerLabwareMovementStrategy.AUTOMATIC
    )

    fs_module_substate = FlexStackerSubState(
        module_id=stacker_id,
        pool_primary_definition=flex_50uL_tiprack,
        pool_adapter_definition=None,
        pool_lid_definition=None,
        contained_labware_bottom_first=_contained_labware(3),
        max_pool_count=3,
        pool_overlap=0,
        pool_height=0,
    )
    decoy.when(
        state_view.modules.get_flex_stacker_substate(module_id=stacker_id)
    ).then_return(fs_module_substate)
    with pytest.raises(
        CannotPerformModuleAction,
        match="Cannot store labware in Flex Stacker in .* because it is full",
    ):
        await subject.execute(data)


async def test_store_raises_if_carriage_logically_empty(
    decoy: Decoy,
    equipment: EquipmentHandler,
    state_view: StateView,
    subject: StoreImpl,
    stacker_id: FlexStackerId,
    flex_50uL_tiprack: LabwareDefinition,
) -> None:
    """It should raise if called with a known-empty carriage."""
    data = flex_stacker.StoreParams(
        moduleId=stacker_id, strategy=StackerLabwareMovementStrategy.AUTOMATIC
    )

    fs_module_substate = FlexStackerSubState(
        module_id=stacker_id,
        pool_primary_definition=flex_50uL_tiprack,
        pool_adapter_definition=None,
        pool_lid_definition=None,
        contained_labware_bottom_first=_contained_labware(1),
        max_pool_count=5,
        pool_overlap=0,
        pool_height=0,
    )
    decoy.when(
        state_view.modules.get_flex_stacker_substate(module_id=stacker_id)
    ).then_return(fs_module_substate)
    decoy.when(state_view.labware.get_id_by_module(stacker_id)).then_raise(
        LabwareNotLoadedOnModuleError()
    )
    with pytest.raises(
        CannotPerformModuleAction,
        match="Flex Stacker in .* cannot store labware because its carriage is empty",
    ):
        await subject.execute(data)


@pytest.mark.parametrize(
    "contained_labware_count,max_pool_count",
    [
        (0, 0),
        (1, 0),
        (0, 1),
    ],
)
async def test_store_raises_if_not_configured(
    contained_labware_count: int,
    max_pool_count: int,
    decoy: Decoy,
    equipment: EquipmentHandler,
    state_view: StateView,
    subject: StoreImpl,
    stacker_id: FlexStackerId,
) -> None:
    """It should raise if called before the stacker is configured."""
    data = flex_stacker.StoreParams(
        moduleId=stacker_id, strategy=StackerLabwareMovementStrategy.AUTOMATIC
    )
    fs_module_substate = FlexStackerSubState(
        module_id=stacker_id,
        pool_primary_definition=None,
        pool_adapter_definition=None,
        pool_lid_definition=None,
        contained_labware_bottom_first=_contained_labware(contained_labware_count),
        max_pool_count=max_pool_count,
        pool_overlap=0,
        pool_height=0,
    )
    decoy.when(
        state_view.modules.get_flex_stacker_substate(module_id=stacker_id)
    ).then_return(fs_module_substate)
    with pytest.raises(
        FlexStackerLabwarePoolNotYetDefinedError,
        match="The Flex Stacker in .* has not been configured yet and cannot be filled.",
    ):
        await subject.execute(data)


@pytest.mark.parametrize(
    "shared_data_error,protocol_engine_error",
    [
        (
            FlexStackerStallError(serial="123", axis=StackerAxis.Z),
            FlexStackerStallOrCollisionError,
        ),
        (
            FlexStackerShuttleMissingError(
                serial="123",
                expected_state=PlatformState.EXTENDED,
                shuttle_state=PlatformState.UNKNOWN,
            ),
            FlexStackerShuttleError,
        ),
        (
            FlexStackerShuttleLabwareError(
                serial="123",
                shuttle_state=PlatformState.UNKNOWN,
                labware_expected=True,
            ),
            FlexStackerLabwareStoreError,
        ),
    ],
)
async def test_store_raises_if_stall(
    decoy: Decoy,
    equipment: EquipmentHandler,
    state_view: StateView,
    subject: StoreImpl,
    model_utils: ModelUtils,
    stacker_id: FlexStackerId,
    flex_50uL_tiprack: LabwareDefinition,
    stacker_hardware: FlexStacker,
    shared_data_error: Exception,
    protocol_engine_error: Type[
        Union[FlexStackerStallOrCollisionError, FlexStackerShuttleError]
    ],
) -> None:
    """It should raise a stall error."""
    data = flex_stacker.StoreParams(
        moduleId=stacker_id, strategy=StackerLabwareMovementStrategy.AUTOMATIC
    )
    error_id = "error-id"
    error_timestamp = datetime(year=2020, month=1, day=2)

    fs_module_substate = FlexStackerSubState(
        module_id=stacker_id,
        pool_primary_definition=flex_50uL_tiprack,
        pool_adapter_definition=None,
        pool_lid_definition=None,
        contained_labware_bottom_first=[],
        max_pool_count=999,
        pool_overlap=6,
        pool_height=10,
    )

    decoy.when(
        state_view.modules.get_flex_stacker_substate(module_id=stacker_id)
    ).then_return(fs_module_substate)

    decoy.when(state_view.labware.get_id_by_module(module_id=stacker_id)).then_return(
        "labware-id"
    )
    decoy.when(
        state_view.labware.get_labware_stack_from_parent("labware-id")
    ).then_return(["labware-id"])
    decoy.when(state_view.labware.get_definition("labware-id")).then_return(
        flex_50uL_tiprack
    )

    decoy.when(state_view.geometry.get_location_sequence("labware-id")).then_return(
        [
            OnModuleLocationSequenceComponent(moduleId=stacker_id),
            OnAddressableAreaLocationSequenceComponent(
                addressableAreaName="flexStackerV1B4",
            ),
            OnCutoutFixtureLocationSequenceComponent(
                cutoutId="cutoutA3", possibleCutoutFixtureIds=["flexStackerModuleV1"]
            ),
        ]
    )
    decoy.when(
        state_view.geometry.get_predicted_location_sequence(
            ModuleLocation(moduleId=stacker_id)
        )
    ).then_return(
        [
            OnModuleLocationSequenceComponent(moduleId=stacker_id),
            OnAddressableAreaLocationSequenceComponent(
                addressableAreaName="flexStackerV1B4",
            ),
            OnCutoutFixtureLocationSequenceComponent(
                cutoutId="cutoutA3", possibleCutoutFixtureIds=["flexStackerModuleV1"]
            ),
        ]
    )

    decoy.when(model_utils.generate_id()).then_return(error_id)
    decoy.when(model_utils.get_timestamp()).then_return(error_timestamp)

    decoy.when(await stacker_hardware.store_labware(labware_height=4)).then_raise(  # type: ignore[func-returns-value]
        shared_data_error
    )

    result = await subject.execute(data)

    assert result == DefinedErrorData(
        public=protocol_engine_error.model_construct(
            id=error_id,
            createdAt=error_timestamp,
            wrappedErrors=[matchers.Anything()],
            errorInfo={"labwareId": "labware-id"},
        ),
        state_update=StateUpdate(),
    )


@pytest.mark.parametrize(
    argnames=["pool_adapter", "pool_lid", "param_adapter", "param_lid"],
    argvalues=[
        pytest.param(
            sentinel.pool_adapter,
            sentinel.pool_lid,
            None,
            sentinel.pool_lid,
            id="missing-adapter",
        ),
        pytest.param(
            sentinel.pool_adapter, None, None, None, id="missing-adapter-no-lid"
        ),
        pytest.param(
            sentinel.pool_adapter,
            sentinel.pool_lid,
            sentinel.pool_adapter,
            None,
            id="missing-lid",
        ),
        pytest.param(None, sentinel.pool_lid, None, None, id="missing-lid-no-adapter"),
        pytest.param(
            None, None, sentinel.param_adapter, None, id="extra-adapter-no-lid"
        ),
        pytest.param(
            None,
            sentinel.pool_lid,
            sentinel.param_adapter,
            sentinel.pool_lid,
            id="extra-adapter",
        ),
        pytest.param(
            sentinel.pool_adapter,
            None,
            sentinel.pool_adapter,
            sentinel.param_lid,
            id="extra-lid",
        ),
        pytest.param(None, None, None, sentinel.param_lid, id="extra-lid-no-adapter"),
        pytest.param(
            None,
            None,
            sentinel.param_adapter,
            sentinel.param_lid,
            id="extra-lid-and-adapter",
        ),
        pytest.param(
            sentinel.pool_adapter,
            sentinel.pool_lid,
            None,
            None,
            id="missing-lid-and-adapter",
        ),
        pytest.param(
            sentinel.pool_adapter,
            sentinel.pool_lid,
            sentinel.pool_adapter,
            sentinel.param_lid,
            id="wrong-lid",
        ),
        pytest.param(
            None, sentinel.pool_lid, None, sentinel.param_lid, id="wrong-lid-no-adapter"
        ),
        pytest.param(
            sentinel.pool_adapter,
            sentinel.pool_lid,
            sentinel.param_adapter,
            sentinel.pool_lid,
            id="wrong-adapter",
        ),
        pytest.param(
            sentinel.pool_adapter,
            None,
            sentinel.param_adapter,
            None,
            id="wrong-adapter-no-lid",
        ),
        pytest.param(
            sentinel.pool_adapter,
            sentinel.pool_lid,
            sentinel.param_adapter,
            sentinel.param_lid,
            id="wrong-lid-and-adapter",
        ),
    ],
)
async def test_store_raises_if_labware_does_not_match(
    decoy: Decoy,
    state_view: StateView,
    equipment: EquipmentHandler,
    subject: StoreImpl,
    stacker_id: FlexStackerId,
    stacker_hardware: FlexStacker,
    pool_adapter: LabwareDefinition | None,
    pool_lid: LabwareDefinition | None,
    param_adapter: LabwareDefinition | None,
    param_lid: LabwareDefinition | None,
) -> None:
    """It should raise if the labware to be stored does not match the labware pool parameters."""
    data = flex_stacker.StoreParams(
        moduleId=stacker_id, strategy=StackerLabwareMovementStrategy.AUTOMATIC
    )

    fs_module_substate = FlexStackerSubState(
        module_id=stacker_id,
        pool_primary_definition=sentinel.primary,
        pool_adapter_definition=pool_adapter,
        pool_lid_definition=pool_lid,
        contained_labware_bottom_first=[],
        max_pool_count=5,
        pool_overlap=0,
        pool_height=0,
    )

    decoy.when(
        state_view.modules.get_flex_stacker_substate(module_id=stacker_id)
    ).then_return(fs_module_substate)

    if param_adapter is not None:
        decoy.when(
            state_view.labware.get_id_by_module(module_id=stacker_id)
        ).then_return("adapter-id")
        if param_lid is not None:
            decoy.when(
                state_view.labware.get_labware_stack_from_parent("adapter-id")
            ).then_return(["lid-id", "labware-id", "adapter-id"])
        else:
            decoy.when(
                state_view.labware.get_labware_stack_from_parent("adapter-id")
            ).then_return(["labware-id", "adapter-id"])
    else:
        decoy.when(
            state_view.labware.get_id_by_module(module_id=stacker_id)
        ).then_return("labware-id")
        if param_lid is not None:
            decoy.when(
                state_view.labware.get_labware_stack_from_parent("labware-id")
            ).then_return(["lid-id", "labware-id"])
        else:
            decoy.when(
                state_view.labware.get_labware_stack_from_parent("labware-id")
            ).then_return(["labware-id"])

    decoy.when(state_view.labware.get_definition("labware-id")).then_return(
        sentinel.primary
    )
    if param_lid is not None:
        decoy.when(state_view.labware.get_definition("lid-id")).then_return(param_lid)
    if param_adapter is not None:
        decoy.when(state_view.labware.get_definition("adapter-id")).then_return(
            param_adapter
        )

    with pytest.raises(
        CannotPerformModuleAction,
        match="Cannot store labware stack that does not correspond with the configuration of Flex Stacker",
    ):
        await subject.execute(data)


@pytest.mark.parametrize(
    "move_strategy",
    [StackerLabwareMovementStrategy.AUTOMATIC, StackerLabwareMovementStrategy.MANUAL],
)
async def test_store(
    decoy: Decoy,
    state_view: StateView,
    equipment: EquipmentHandler,
    stacker_id: FlexStackerId,
    subject: StoreImpl,
    stacker_hardware: FlexStacker,
    flex_50uL_tiprack: LabwareDefinition,
    move_strategy: StackerLabwareMovementStrategy,
) -> None:
    """It should store the labware on the stack."""
    data = flex_stacker.StoreParams(moduleId=stacker_id, strategy=move_strategy)

    fs_module_substate = FlexStackerSubState(
        module_id=stacker_id,
        pool_primary_definition=flex_50uL_tiprack,
        pool_adapter_definition=None,
        pool_lid_definition=None,
        contained_labware_bottom_first=_contained_labware(1),
        max_pool_count=5,
        pool_overlap=6,
        pool_height=10,
    )

    decoy.when(
        state_view.modules.get_flex_stacker_substate(module_id=stacker_id)
    ).then_return(fs_module_substate)

    decoy.when(state_view.labware.get_id_by_module(module_id=stacker_id)).then_return(
        "labware-id"
    )
    decoy.when(
        state_view.labware.get_labware_stack_from_parent("labware-id")
    ).then_return(["labware-id"])
    decoy.when(state_view.labware.get_definition("labware-id")).then_return(
        flex_50uL_tiprack
    )
    decoy.when(state_view.geometry.get_location_sequence("labware-id")).then_return(
        [
            OnModuleLocationSequenceComponent(moduleId=stacker_id),
            OnAddressableAreaLocationSequenceComponent(
                addressableAreaName="flexStackerV1B4",
            ),
            OnCutoutFixtureLocationSequenceComponent(
                cutoutId="cutoutA3", possibleCutoutFixtureIds=["flexStackerModuleV1"]
            ),
        ]
    )
    decoy.when(
        state_view.geometry.get_predicted_location_sequence(
            ModuleLocation(moduleId=stacker_id)
        )
    ).then_return(
        [
            OnModuleLocationSequenceComponent(moduleId=stacker_id),
            OnAddressableAreaLocationSequenceComponent(
                addressableAreaName="flexStackerV1B4",
            ),
            OnCutoutFixtureLocationSequenceComponent(
                cutoutId="cutoutA3", possibleCutoutFixtureIds=["flexStackerModuleV1"]
            ),
        ]
    )

    decoy.when(
        state_view.labware.get_uri_from_definition_unless_none(flex_50uL_tiprack)
    ).then_return("opentrons/opentrons_flex_96_filtertiprack_50ul/1")

    result = await subject.execute(data)

    decoy.verify(
        await stacker_hardware.store_labware(labware_height=4),
        times=1 if move_strategy == StackerLabwareMovementStrategy.AUTOMATIC else 0,
    )

    assert result == SuccessData(
        public=flex_stacker.StoreResult(
            primaryOriginLocationSequence=[
                OnModuleLocationSequenceComponent(moduleId=stacker_id),
                OnAddressableAreaLocationSequenceComponent(
                    addressableAreaName="flexStackerV1B4",
                ),
                OnCutoutFixtureLocationSequenceComponent(
                    cutoutId="cutoutA3",
                    possibleCutoutFixtureIds=["flexStackerModuleV1"],
                ),
            ],
            primaryLabwareId="labware-id",
            eventualDestinationLocationSequence=[
                InStackerHopperLocation(moduleId=stacker_id),
            ],
            primaryLabwareURI="opentrons/opentrons_flex_96_filtertiprack_50ul/1",
        ),
        state_update=StateUpdate(
            batch_labware_location=BatchLabwareLocationUpdate(
                new_locations_by_id={
                    "labware-id": InStackerHopperLocation(moduleId=stacker_id)
                },
                new_offset_ids_by_id={"labware-id": None},
            ),
            flex_stacker_state_update=FlexStackerStateUpdate(
                module_id=stacker_id,
                contained_labware_bottom_first=(
                    [
                        StackerStoredLabwareGroup(
                            primaryLabwareId="labware-id",
                            adapterLabwareId=None,
                            lidLabwareId=None,
                        ),
                    ]
                    + _contained_labware(1)
                ),
            ),
        ),
    )



"""Unit tests for `opentrons.util.broker`."""

from typing import List

from opentrons.util.broker import Broker


def test_context_manager_subscriptions() -> None:
    """Test subscribing, receiving messages, and unsubscribing."""
    subject = Broker[int]()

    received_by_a: List[int] = []
    received_by_b: List[int] = []

    def callback_a(message: int) -> None:
        received_by_a.append(message)

    def callback_b(message: int) -> None:
        received_by_b.append(message)

    subject.publish(1)
    with subject.subscribed(callback_a):
        subject.publish(2)
        with subject.subscribed(callback_b):
            subject.publish(3)
        subject.publish(4)
    subject.publish(5)

    assert received_by_a == [2, 3, 4]
    assert received_by_b == [3]


def test_non_context_manager_subscriptions() -> None:
    """Test subscribing, receiving messages, and unsubscribing."""
    subject = Broker[int]()

    received_by_a: List[int] = []
    received_by_b: List[int] = []

    def callback_a(message: int) -> None:
        received_by_a.append(message)

    def callback_b(message: int) -> None:
        received_by_b.append(message)

    subject.publish(1)
    unsubscribe_a = subject.subscribe(callback_a)
    subject.publish(2)
    unsubscribe_b = subject.subscribe(callback_b)
    subject.publish(3)
    unsubscribe_b()
    subject.publish(4)
    unsubscribe_a()
    subject.publish(5)

    assert received_by_a == [2, 3, 4]
    assert received_by_b == [3]



"""Tests for the legacy Protocol API module core implementations."""

import pytest
from decoy import Decoy

from opentrons.hardware_control import SynchronousAdapter
from opentrons.hardware_control.modules import MagDeck, MagneticStatus
from opentrons.hardware_control.modules.types import MagneticModuleModel
from opentrons.protocol_api.core.legacy.legacy_module_core import (
    LegacyMagneticModuleCore,
    create_module_core,
)
from opentrons.protocol_api.core.legacy.legacy_protocol_core import (
    LegacyProtocolCore,
)
from opentrons.protocol_api.core.legacy.module_geometry import ModuleGeometry


@pytest.fixture
def mock_geometry(decoy: Decoy) -> ModuleGeometry:
    """Get a mock module geometry."""
    return decoy.mock(cls=ModuleGeometry)


@pytest.fixture
def mock_sync_module_hardware(decoy: Decoy) -> SynchronousAdapter[MagDeck]:
    """Get a mock module hardware control interface."""
    return decoy.mock(name="SynchronousAdapter[AbstractModule]")  # type: ignore[no-any-return]


@pytest.fixture
def mock_protocol_core(decoy: Decoy) -> LegacyProtocolCore:
    """Get a mock protocol core."""
    return decoy.mock(cls=LegacyProtocolCore)


@pytest.fixture
def subject(
    mock_geometry: ModuleGeometry,
    mock_sync_module_hardware: SynchronousAdapter[MagDeck],
    mock_protocol_core: LegacyProtocolCore,
) -> LegacyMagneticModuleCore:
    """Get a legacy module implementation core with mocked out dependencies."""
    return LegacyMagneticModuleCore(
        requested_model=MagneticModuleModel.MAGNETIC_V1,
        geometry=mock_geometry,
        sync_module_hardware=mock_sync_module_hardware,
        protocol_core=mock_protocol_core,
    )


def test_create(
    decoy: Decoy,
    mock_geometry: ModuleGeometry,
    mock_protocol_core: LegacyProtocolCore,
) -> None:
    """It should be able to create a magnetic module core."""
    mock_module_hardware_api = decoy.mock(cls=MagDeck)
    result = create_module_core(
        geometry=mock_geometry,
        module_hardware_api=mock_module_hardware_api,
        requested_model=MagneticModuleModel.MAGNETIC_V1,
        protocol_core=mock_protocol_core,
    )

    assert isinstance(result, LegacyMagneticModuleCore)


def test_disengage(
    decoy: Decoy,
    mock_sync_module_hardware: SynchronousAdapter[MagDeck],
    subject: LegacyMagneticModuleCore,
) -> None:
    """It should disengage by calling the hardware API."""
    subject.disengage()

    decoy.verify(mock_sync_module_hardware.deactivate(), times=1)


def test_get_status(
    decoy: Decoy,
    mock_sync_module_hardware: SynchronousAdapter[MagDeck],
    subject: LegacyMagneticModuleCore,
) -> None:
    """It should report the hardware status."""
    decoy.when(mock_sync_module_hardware.status).then_return(MagneticStatus.DISENGAGED)
    result = subject.get_status()
    assert result == MagneticStatus.DISENGAGED


def test_engage_height_from_home(
    decoy: Decoy,
    mock_sync_module_hardware: SynchronousAdapter[MagDeck],
    subject: LegacyMagneticModuleCore,
) -> None:
    """It should engage height from home."""
    subject.engage(height_from_home=42.0)

    decoy.verify(mock_sync_module_hardware.engage(height=42.0), times=1)


def test_engage_height_from_base(
    decoy: Decoy,
    mock_sync_module_hardware: SynchronousAdapter[MagDeck],
    subject: LegacyMagneticModuleCore,
) -> None:
    """It should engage height from base."""
    subject.engage(height_from_base=42.0)

    decoy.verify(mock_sync_module_hardware.engage(height_from_base=42.0), times=1)


def test_engage_height_from_labware(
    decoy: Decoy,
    mock_geometry: ModuleGeometry,
    mock_sync_module_hardware: SynchronousAdapter[MagDeck],
    subject: LegacyMagneticModuleCore,
) -> None:
    """It should engage height from the labware's defined position."""
    decoy.when(
        mock_geometry.labware._core.get_default_magnet_engage_height(False)  # type: ignore[union-attr]
    ).then_return(32.0)

    subject.engage_to_labware(offset=10.0)

    decoy.verify(mock_sync_module_hardware.engage(height=42.0), times=1)


def test_engage_height_from_labware_no_labware(
    decoy: Decoy,
    mock_geometry: ModuleGeometry,
    subject: LegacyMagneticModuleCore,
) -> None:
    """It should raise if no labware is loaded."""
    decoy.when(mock_geometry.labware).then_return(None)

    with pytest.raises(ValueError):
        subject.engage_to_labware()


def test_engage_height_from_labware_no_engage_height(
    decoy: Decoy,
    mock_geometry: ModuleGeometry,
    subject: LegacyMagneticModuleCore,
) -> None:
    """It should raise if the labware is not magnetic module compatible."""
    decoy.when(
        mock_geometry.labware._core.get_default_magnet_engage_height(),  # type: ignore[union-attr]
        ignore_extra_args=True,
    ).then_return(None)

    with pytest.raises(ValueError):
        subject.engage_to_labware()


def test_engage_height_from_labware_with_gen1(
    decoy: Decoy,
    mock_geometry: ModuleGeometry,
    mock_sync_module_hardware: SynchronousAdapter[MagDeck],
    subject: LegacyMagneticModuleCore,
) -> None:
    """It should double the labware's height if it's a GEN1 magdeck."""
    decoy.when(mock_geometry.model).then_return(MagneticModuleModel.MAGNETIC_V1)

    decoy.when(
        mock_geometry.labware._core.get_default_magnet_engage_height(False)  # type: ignore[union-attr]
    ).then_return(32.0)

    subject.engage_to_labware(offset=10.0)

    decoy.verify(mock_sync_module_hardware.engage(height=74.0), times=1)


def test_engage_height_from_labware_with_gen1_preserve_half_mm(
    decoy: Decoy,
    mock_geometry: ModuleGeometry,
    mock_sync_module_hardware: SynchronousAdapter[MagDeck],
    subject: LegacyMagneticModuleCore,
) -> None:
    """It should not double the labware's height if told to preserve half-mm."""
    decoy.when(mock_geometry.model).then_return(MagneticModuleModel.MAGNETIC_V1)

    decoy.when(
        mock_geometry.labware._core.get_default_magnet_engage_height(True)  # type: ignore[union-attr]
    ).then_return(32.0)

    subject.engage_to_labware(offset=10.0, preserve_half_mm=True)

    decoy.verify(mock_sync_module_hardware.engage(height=42.0), times=1)



"""StartRunProfile command request, result, and implementation models."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, List, Optional, Union, overload

from pydantic import BaseModel, Field
from pydantic.json_schema import SkipJsonSchema
from typing_extensions import Literal, Type

from ...errors.error_occurrence import ErrorOccurrence
from ..command import AbstractCommandImpl, BaseCommand, BaseCommandCreate, SuccessData
from .run_extended_profile import ProfileCycle, ProfileStep
from opentrons.hardware_control.modules.types import ThermocyclerCycle, ThermocyclerStep

if TYPE_CHECKING:
    from opentrons.protocol_engine.execution import (
        EquipmentHandler,
        TaskHandler,
    )
    from opentrons.protocol_engine.state.module_substates.thermocycler_module_substate import (
        ThermocyclerModuleSubState,
    )
    from opentrons.protocol_engine.state.state import StateView

StartRunExtendedProfileCommandType = Literal["thermocycler/startRunExtendedProfile"]


def _remove_default(s: dict[str, Any]) -> None:
    s.pop("default", None)


class StartRunExtendedProfileStepParams(BaseModel):
    """Input parameters for an individual Thermocycler profile step."""

    celsius: float = Field(..., description="Target temperature in °C.")
    holdSeconds: float = Field(
        ..., description="Time to hold target temperature at in seconds."
    )
    rampRate: float | SkipJsonSchema[None] = Field(
        None,
        description="How quickly to change temperature in °C/second.",
        json_schema_extra=_remove_default,
    )


class StartRunExtendedProfileParams(BaseModel):
    """Input parameters to run a Thermocycler profile."""

    moduleId: str = Field(..., description="Unique ID of the Thermocycler.")
    profileElements: List[Union[ProfileStep, ProfileCycle]] = Field(
        ...,
        description="Elements of the profile. Each can be either a step or a cycle.",
    )
    blockMaxVolumeUl: float | SkipJsonSchema[None] = Field(
        None,
        description="Amount of liquid in uL of the most-full well"
        " in labware loaded onto the thermocycler.",
        json_schema_extra=_remove_default,
    )
    taskId: str | None = Field(None, description="The id of the profile task")


class StartRunExtendedProfileResult(BaseModel):
    """Result data from running a Thermocycler profile."""

    taskId: str = Field(..., description="The id of the profile task")


def _transform_profile_step(
    step: ProfileStep, thermocycler_state: ThermocyclerModuleSubState
) -> ThermocyclerStep:
    return ThermocyclerStep(
        temperature=thermocycler_state.validate_target_block_temperature(step.celsius),
        hold_time_seconds=step.holdSeconds,
        ramp_rate=thermocycler_state.validate_ramp_rate(step.rampRate, step.celsius),
    )


@overload
def _transform_profile_element(
    element: ProfileStep, thermocycler_state: ThermocyclerModuleSubState
) -> ThermocyclerStep: ...


@overload
def _transform_profile_element(
    element: ProfileCycle, thermocycler_state: ThermocyclerModuleSubState
) -> ThermocyclerCycle: ...


def _transform_profile_element(
    element: Union[ProfileStep, ProfileCycle],
    thermocycler_state: ThermocyclerModuleSubState,
) -> Union[ThermocyclerStep, ThermocyclerCycle]:
    if isinstance(element, ProfileStep):
        return _transform_profile_step(element, thermocycler_state)
    else:
        return ThermocyclerCycle(
            steps=[
                _transform_profile_step(step, thermocycler_state)
                for step in element.steps
            ],
            repetitions=element.repetitions,
        )


class StartRunExtendedProfileImpl(
    AbstractCommandImpl[
        StartRunExtendedProfileParams, SuccessData[StartRunExtendedProfileResult]
    ]
):
    """Execution implementation of a Thermocycler's run profile command."""

    def __init__(
        self,
        state_view: StateView,
        equipment: EquipmentHandler,
        task_handler: TaskHandler,
        **unused_dependencies: object,
    ) -> None:
        self._state_view = state_view
        self._equipment = equipment
        self._task_handler = task_handler

    async def execute(
        self, params: StartRunExtendedProfileParams
    ) -> SuccessData[StartRunExtendedProfileResult]:
        """Run a Thermocycler profile."""
        thermocycler_state = self._state_view.modules.get_thermocycler_module_substate(
            params.moduleId
        )
        thermocycler_hardware = self._equipment.get_module_hardware_api(
            thermocycler_state.module_id
        )

        profile = [
            _transform_profile_element(element, thermocycler_state)
            for element in params.profileElements
        ]
        target_volume: Optional[float]
        if params.blockMaxVolumeUl is not None:
            target_volume = thermocycler_state.validate_max_block_volume(
                params.blockMaxVolumeUl
            )
        else:
            target_volume = None

        async def start_run_profile(task_handler: TaskHandler) -> None:
            if thermocycler_hardware is not None:
                async with task_handler.synchronize_cancel_latest(
                    thermocycler_state.module_id + "-block"
                ):
                    await thermocycler_hardware.execute_profile(
                        profile=profile, volume=target_volume
                    )

        task = await self._task_handler.create_task(
            task_function=start_run_profile, id=params.taskId
        )

        return SuccessData(
            public=StartRunExtendedProfileResult(taskId=task.id),
        )


class StartRunExtendedProfile(
    BaseCommand[
        StartRunExtendedProfileParams, StartRunExtendedProfileResult, ErrorOccurrence
    ]
):
    """A command to start the execution of a Thermocycler profile run without waiting for its completion."""

    commandType: StartRunExtendedProfileCommandType = (
        "thermocycler/startRunExtendedProfile"
    )
    params: StartRunExtendedProfileParams

    _ImplementationCls: Type[StartRunExtendedProfileImpl] = StartRunExtendedProfileImpl


class StartRunExtendedProfileCreate(BaseCommandCreate[StartRunExtendedProfileParams]):
    """A request to execute a Thermocycler profile run."""

    commandType: StartRunExtendedProfileCommandType = (
        "thermocycler/startRunExtendedProfile"
    )
    params: StartRunExtendedProfileParams

    _CommandCls: Type[StartRunExtendedProfile] = StartRunExtendedProfile



from typing import Any, Dict, cast

import pytest
from _pytest.fixtures import SubRequest
from pytest_lazy_fixtures import lf as lazy_fixture

from opentrons.config.advanced_settings import _ensure, _migrate


@pytest.fixture
def migrated_file_version() -> int:
    return 39


# make sure to set a boolean value in default_file_settings only if
# that value has a migration that changes an old default value to a new one
@pytest.fixture
def default_file_settings() -> Dict[str, Any]:
    return {
        "shortFixedTrash": None,
        "deckCalibrationDots": None,
        "disableHomeOnBoot": None,
        "useOldAspirationFunctions": None,
        "enableDoorSafetySwitch": None,
        "enableOT3HardwareController": None,
        "rearPanelIntegration": True,
        "disableStallDetection": None,
        "disableStatusBar": None,
        "disableOverpressureDetection": None,
        "estopNotRequired": None,
        "enableErrorRecoveryExperiments": None,
        "enableOEMMode": None,
        "enablePerformanceMetrics": None,
        "disableFlexStackerLabwareDetection": None,
        "enableProtocolSubprocess": False,
        "enableHardwareSubprocess": False,
    }


@pytest.fixture
def empty_settings() -> Dict[str, Any]:
    return {}


@pytest.fixture
def version_less() -> Dict[str, Any]:
    return {
        "shortFixedTrash": True,
        "calibrateToBottom": True,
        "deckCalibrationDots": True,
        "disableHomeOnBoot": True,
        "useOldAspirationFunctions": True,
    }


@pytest.fixture
def v1_config() -> Dict[str, Any]:
    return {
        "_version": 1,
        "shortFixedTrash": True,
        "calibrateToBottom": True,
        "deckCalibrationDots": True,
        "disableHomeOnBoot": True,
        "useProtocolApi2": None,
        "useOldAspirationFunctions": True,
    }


@pytest.fixture
def v2_config(v1_config: Dict[str, Any]) -> Dict[str, Any]:
    r = v1_config.copy()
    r.update(
        {
            "_version": 2,
            "disableLogAggregation": None,
        }
    )
    return r


@pytest.fixture
def v3_config(v2_config: Dict[str, Any]) -> Dict[str, Any]:
    r = v2_config.copy()
    r.update({"_version": 3, "enableApi1BackCompat": False})
    return r


@pytest.fixture
def v4_config(v3_config: Dict[str, Any]) -> Dict[str, Any]:
    r = v3_config.copy()
    r.update({"_version": 4, "useV1HttpApi": False})
    return r


@pytest.fixture
def v5_config(v4_config: Dict[str, Any]) -> Dict[str, Any]:
    r = v4_config.copy()
    r.update(
        {
            "_version": 5,
            "enableDoorSafetySwitch": True,
        }
    )
    return r


@pytest.fixture
def v6_config(v5_config: Dict[str, Any]) -> Dict[str, Any]:
    r = v5_config.copy()
    r.update({"_version": 6, "enableTipLengthCalibration": True})
    return r


@pytest.fixture
def v7_config(v6_config: Dict[str, Any]) -> Dict[str, Any]:
    r = v6_config.copy()
    r.update(
        {
            "_version": 7,
            "enableHttpProtocolSessions": True,
        }
    )
    return r


@pytest.fixture
def v8_config(v7_config: Dict[str, Any]) -> Dict[str, Any]:
    r = v7_config.copy()
    r.update(
        {
            "_version": 8,
            "enableFastProtocolUpload": True,
        }
    )
    return r


@pytest.fixture
def v9_config(v8_config: Dict[str, Any]) -> Dict[str, Any]:
    r = v8_config.copy()
    r.update(
        {
            "_version": 9,
            "enableProtocolEngine": True,
        }
    )
    return r


@pytest.fixture
def v10_config(v9_config: Dict[str, Any]) -> Dict[str, Any]:
    r = v9_config.copy()
    r.pop("useProtocolApi2")
    r.pop("enableApi1BackCompat")
    r.pop("useV1HttpApi")
    r.pop("enableTipLengthCalibration")
    r.pop("enableFastProtocolUpload")
    r.update(
        {
            "_version": 10,
            "disableFastProtocolUpload": True,
        }
    )
    return r


@pytest.fixture
def v11_config(v10_config: Dict[str, Any]) -> Dict[str, Any]:
    r = v10_config.copy()
    r.pop("enableProtocolEngine")
    r.update({"_version": 11})
    return r


@pytest.fixture
def v12_config(v11_config: Dict[str, Any]) -> Dict[str, Any]:
    r = v11_config.copy()
    r.update(
        {
            "_version": 12,
            "enableOT3HardwareController": True,
        }
    )
    return r


@pytest.fixture
def v13_config(v12_config: Dict[str, Any]) -> Dict[str, Any]:
    r = v12_config.copy()
    r.pop("calibrateToBottom")
    r.update({"_version": 13})
    return r


@pytest.fixture
def v14_config(v13_config: Dict[str, Any]) -> Dict[str, Any]:
    r = v13_config.copy()
    r.pop("enableHttpProtocolSessions")
    r.update({"_version": 14})
    return r


@pytest.fixture
def v15_config(v14_config: Dict[str, Any]) -> Dict[str, Any]:
    r = v14_config.copy()
    r.update(
        {
            "_version": 15,
            "enableHeaterShakerPAPI": True,
        }
    )
    return r


@pytest.fixture
def v16_config(v15_config: Dict[str, Any]) -> Dict[str, Any]:
    r = v15_config.copy()
    r.pop("enableHeaterShakerPAPI")
    r.update({"_version": 16})
    return r


@pytest.fixture
def v17_config(v16_config: Dict[str, Any]) -> Dict[str, Any]:
    r = v16_config.copy()
    r.update(
        {
            "_version": 17,
            "enableProtocolEnginePAPICore": True,
        }
    )
    return r


@pytest.fixture
def v18_config(v17_config: Dict[str, Any]) -> Dict[str, Any]:
    r = v17_config.copy()
    r.update(
        {
            "_version": 18,
            "enableLoadLiquid": True,
        }
    )
    return r


@pytest.fixture
def v19_config(v18_config: Dict[str, Any]) -> Dict[str, Any]:
    r = v18_config.copy()
    r.pop("enableLoadLiquid")
    r.update({"_version": 19})
    return r


@pytest.fixture
def v20_config(v19_config: Dict[str, Any]) -> Dict[str, Any]:
    r = v19_config.copy()
    r.update(
        {
            "_version": 20,
            "enableOT3FirmwareUpdates": None,
        }
    )
    return r


@pytest.fixture
def v21_config(v20_config: Dict[str, Any]) -> Dict[str, Any]:
    r = v20_config.copy()
    r.pop("enableProtocolEnginePAPICore")
    r.update({"_version": 21})
    return r


@pytest.fixture
def v22_config(v21_config: Dict[str, Any]) -> Dict[str, Any]:
    r = v21_config.copy()
    r.pop("enableOT3FirmwareUpdates")
    r.update({"_version": 22})
    return r


@pytest.fixture
def v23_config(v22_config: Dict[str, Any]) -> Dict[str, Any]:
    r = v22_config.copy()
    r.update(
        {
            "_version": 23,
            "rearPanelIntegration": None,
        }
    )
    return r


@pytest.fixture
def v24_config(v23_config: Dict[str, Any]) -> Dict[str, Any]:
    r = v23_config.copy()
    r.update(
        {
            "_version": 24,
            "rearPanelIntegration": True,
        }
    )
    return r


@pytest.fixture
def v25_config(v24_config: Dict[str, Any]) -> Dict[str, Any]:
    r = v24_config.copy()
    r.update(
        {
            "_version": 25,
            "disableStallDetection": None,
        }
    )
    return r


@pytest.fixture
def v26_config(v25_config: Dict[str, Any]) -> Dict[str, Any]:
    r = v25_config.copy()
    r.update(
        {
            "_version": 26,
            "disableStatusBar": None,
        }
    )
    return r


@pytest.fixture
def v27_config(v26_config: Dict[str, Any]) -> Dict[str, Any]:
    r = v26_config.copy()
    r.update(
        {
            "_version": 27,
            "disableOverpressureDetection": None,
        }
    )
    return r


@pytest.fixture
def v28_config(v27_config: Dict[str, Any]) -> Dict[str, Any]:
    r = v27_config.copy()
    r.update(
        {
            "_version": 28,
            "disableTipPresenceDetection": None,
        }
    )
    return r


@pytest.fixture
def v29_config(v28_config: Dict[str, Any]) -> Dict[str, Any]:
    r = v28_config.copy()
    r.update(
        {
            "_version": 29,
            "estopNotRequired": None,
        }
    )
    return r


@pytest.fixture
def v30_config(v29_config: Dict[str, Any]) -> Dict[str, Any]:
    r = {k: v for k, v in v29_config.items() if k != "disableTipPresenceDetection"}
    r["_version"] = 30
    return r


@pytest.fixture
def v31_config(v30_config: Dict[str, Any]) -> Dict[str, Any]:
    r = v30_config.copy()
    r.update(
        {
            "_version": 31,
            "enableErrorRecoveryExperiments": None,
        }
    )
    return r


@pytest.fixture
def v32_config(v31_config: Dict[str, Any]) -> Dict[str, Any]:
    r = v31_config.copy()
    r.update(
        {
            "_version": 32,
            "enableOEMMode": None,
        }
    )
    return r


@pytest.fixture
def v33_config(v32_config: Dict[str, Any]) -> Dict[str, Any]:
    r = v32_config.copy()
    r.update(
        {
            "_version": 33,
            "enablePerformanceMetrics": None,
        }
    )
    return r


@pytest.fixture
def v34_config(v33_config: Dict[str, Any]) -> Dict[str, Any]:
    r = v33_config.copy()
    r.pop("disableFastProtocolUpload")
    r["_version"] = 34
    return r


@pytest.fixture
def v35_config(v34_config: Dict[str, Any]) -> Dict[str, Any]:
    r = v34_config.copy()
    r.pop("disableLogAggregation")
    r["_version"] = 35
    return r


@pytest.fixture
def v36_config(v35_config: Dict[str, Any]) -> Dict[str, Any]:
    r = v35_config.copy()
    r.update(
        {
            "_version": 36,
            "allowLiquidClasses": None,
        }
    )
    return r


@pytest.fixture
def v37_config(v36_config: Dict[str, Any]) -> Dict[str, Any]:
    r = {k: v for k, v in v36_config.items() if k != "allowLiquidClasses"}
    r["_version"] = 37
    return r


@pytest.fixture
def v38_config(v37_config: Dict[str, Any]) -> Dict[str, Any]:
    r = v37_config.copy()
    r.update(
        {
            "_version": 38,
            "disableFlexStackerLabwareDetection": None,
        }
    )
    return r


@pytest.fixture
def v39_config(v38_config: Dict[str, Any]) -> Dict[str, Any]:
    r = v38_config.copy()
    r.update(
        {
            "_version": 39,
            "enableProtocolSubprocess": False,
            "enableHardwareSubprocess": False,
        }
    )
    return r


@pytest.fixture(
    params=[
        lazy_fixture("empty_settings"),
        lazy_fixture("version_less"),
        lazy_fixture("v1_config"),
        lazy_fixture("v2_config"),
        lazy_fixture("v3_config"),
        lazy_fixture("v4_config"),
        lazy_fixture("v5_config"),
        lazy_fixture("v6_config"),
        lazy_fixture("v7_config"),
        lazy_fixture("v8_config"),
        lazy_fixture("v9_config"),
        lazy_fixture("v10_config"),
        lazy_fixture("v11_config"),
        lazy_fixture("v12_config"),
        lazy_fixture("v13_config"),
        lazy_fixture("v14_config"),
        lazy_fixture("v15_config"),
        lazy_fixture("v16_config"),
        lazy_fixture("v17_config"),
        lazy_fixture("v18_config"),
        lazy_fixture("v19_config"),
        lazy_fixture("v20_config"),
        lazy_fixture("v21_config"),
        lazy_fixture("v22_config"),
        lazy_fixture("v23_config"),
        lazy_fixture("v24_config"),
        lazy_fixture("v25_config"),
        lazy_fixture("v26_config"),
        lazy_fixture("v27_config"),
        lazy_fixture("v28_config"),
        lazy_fixture("v29_config"),
        lazy_fixture("v30_config"),
        lazy_fixture("v31_config"),
        lazy_fixture("v32_config"),
        lazy_fixture("v33_config"),
        lazy_fixture("v34_config"),
        lazy_fixture("v35_config"),
        lazy_fixture("v36_config"),
        lazy_fixture("v37_config"),
        lazy_fixture("v38_config"),
        lazy_fixture("v39_config"),
    ],
)
def old_settings(request: SubRequest) -> Dict[str, Any]:
    return cast(Dict[str, Any], request.param)


def test_migrations(
    old_settings: Dict[str, Any],
    migrated_file_version: int,
    default_file_settings: Dict[str, Any],
) -> None:
    settings, version = _migrate(old_settings)

    expected = default_file_settings.copy()
    # this updates the default file settings value to the previous default
    # the final check if default_file_settings[k] is None was added for when
    # a migration changes the default value. Without it we overwrite the new
    # default value with the old default value
    expected.update(
        {
            k: v
            for k, v in old_settings.items()
            if k != "_version"
            and k in default_file_settings
            and default_file_settings[k] is None
        }
    )

    assert version == migrated_file_version
    assert settings == expected


def test_migrates_versionless_old_config(
    migrated_file_version: int,
    default_file_settings: Dict[str, Any],
) -> None:
    settings, version = _migrate(
        {
            "short-fixed-trash": False,
            "calibrate-to-bottom": False,
            "dots-deck-type": True,
            "disable-home-on-boot": False,
        }
    )

    expected = default_file_settings.copy()
    expected.update(
        {
            "shortFixedTrash": None,
            "deckCalibrationDots": True,
            "disableHomeOnBoot": None,
        }
    )

    assert version == migrated_file_version
    assert settings == expected


def test_ignores_invalid_keys(
    migrated_file_version: int,
    default_file_settings: Dict[str, Any],
) -> None:
    settings, version = _migrate(
        {
            "split-labware-def": True,
            "splitLabwareDefinitions": True,
        }
    )

    assert version == migrated_file_version
    assert settings == default_file_settings


def test_ensures_config() -> None:
    assert _ensure({"_version": 3, "shortFixedTrash": False}) == {
        "_version": 3,
        "shortFixedTrash": False,
        "deckCalibrationDots": None,
        "disableHomeOnBoot": None,
        "useOldAspirationFunctions": None,
        "enableDoorSafetySwitch": None,
        "enableOT3HardwareController": None,
        "rearPanelIntegration": None,
        "disableStallDetection": None,
        "disableStatusBar": None,
        "estopNotRequired": None,
        "disableOverpressureDetection": None,
        "enableErrorRecoveryExperiments": None,
        "enableOEMMode": None,
        "enablePerformanceMetrics": None,
        "disableFlexStackerLabwareDetection": None,
        "enableProtocolSubprocess": None,
        "enableHardwareSubprocess": None,
    }



"""Smoke tests for the CommandExecutor class."""

import asyncio
from datetime import datetime
from typing import Any, Optional, Type, Union, cast

import pytest
from decoy import Decoy, matchers
from pydantic import BaseModel, PrivateAttr

from opentrons_shared_data.errors.exceptions import EStopActivatedError, PythonException

from opentrons.hardware_control import HardwareControlAPI, OT2HardwareControlAPI
from opentrons.protocol_engine import errors
from opentrons.protocol_engine.actions import (
    ActionDispatcher,
    FailCommandAction,
    RunCommandAction,
    SucceedCommandAction,
)
from opentrons.protocol_engine.commands import (
    AbstractCommandImpl,
    BaseCommand,
    Command,
    CommandStatus,
)
from opentrons.protocol_engine.commands.command import DefinedErrorData, SuccessData
from opentrons.protocol_engine.error_recovery_policy import (
    ErrorRecoveryPolicy,
    ErrorRecoveryType,
    never_recover,
)
from opentrons.protocol_engine.errors.error_occurrence import ErrorOccurrence
from opentrons.protocol_engine.errors.exceptions import (
    EStopActivatedError as PE_EStopActivatedError,
)
from opentrons.protocol_engine.execution import (
    CommandExecutor,
    EquipmentHandler,
    GantryMover,
    LabwareMovementHandler,
    MovementHandler,
    PipettingHandler,
    RailLightsHandler,
    RunControlHandler,
    StatusBarHandler,
    TaskHandler,
    TipHandler,
)
from opentrons.protocol_engine.execution.command_executor import (
    CommandNoteTrackerProvider,
)
from opentrons.protocol_engine.notes import CommandNote, CommandNoteTracker
from opentrons.protocol_engine.resources import CameraProvider, FileProvider, ModelUtils
from opentrons.protocol_engine.resources.camera_provider import (
    CameraSettings,
    ImageParameters,
)
from opentrons.protocol_engine.state.state import StateStore


@pytest.fixture
def hardware_api(decoy: Decoy) -> HardwareControlAPI:
    """Get a mocked out StateStore."""
    return decoy.mock(cls=OT2HardwareControlAPI)


@pytest.fixture
def state_store(decoy: Decoy) -> StateStore:
    """Get a mocked out StateStore."""
    return decoy.mock(cls=StateStore)


@pytest.fixture
def action_dispatcher(decoy: Decoy) -> ActionDispatcher:
    """Get a mocked out ActionDispatcher."""
    return decoy.mock(cls=ActionDispatcher)


@pytest.fixture
def equipment(decoy: Decoy) -> EquipmentHandler:
    """Get a mocked out EquipmentHandler."""
    return decoy.mock(cls=EquipmentHandler)


@pytest.fixture
def movement(decoy: Decoy) -> MovementHandler:
    """Get a mocked out MovementHandler."""
    return decoy.mock(cls=MovementHandler)


@pytest.fixture
def mock_gantry_mover(decoy: Decoy) -> GantryMover:
    """Get a mocked out GantryMover."""
    return decoy.mock(cls=GantryMover)


@pytest.fixture
def labware_movement(decoy: Decoy) -> LabwareMovementHandler:
    """Get a mocked out LabwareMovementHandler."""
    return decoy.mock(cls=LabwareMovementHandler)


@pytest.fixture
def pipetting(decoy: Decoy) -> PipettingHandler:
    """Get a mocked out PipettingHandler."""
    return decoy.mock(cls=PipettingHandler)


@pytest.fixture
def mock_tip_handler(decoy: Decoy) -> TipHandler:
    """Get a mocked out TipHandler."""
    return decoy.mock(cls=TipHandler)


@pytest.fixture
def run_control(decoy: Decoy) -> RunControlHandler:
    """Get a mocked out RunControlHandler."""
    return decoy.mock(cls=RunControlHandler)


@pytest.fixture
def model_utils(decoy: Decoy) -> ModelUtils:
    """Get a mocked out ModelUtils."""
    return decoy.mock(cls=ModelUtils)


@pytest.fixture
def rail_lights(decoy: Decoy) -> RailLightsHandler:
    """Get a mocked out RunControlHandler."""
    return decoy.mock(cls=RailLightsHandler)


@pytest.fixture
def status_bar(decoy: Decoy) -> StatusBarHandler:
    """Get a mocked out StatusBarHandler."""
    return decoy.mock(cls=StatusBarHandler)


@pytest.fixture
def task_handler(decoy: Decoy) -> TaskHandler:
    """Get a mocked out TaskHandler."""
    return decoy.mock(cls=TaskHandler)


@pytest.fixture
def command_note_tracker_provider(decoy: Decoy) -> CommandNoteTrackerProvider:
    """Get a mock tracker provider."""
    return decoy.mock(cls=CommandNoteTrackerProvider)


@pytest.fixture
def error_recovery_policy(state_store: StateStore, decoy: Decoy) -> ErrorRecoveryPolicy:
    """Get a mock error recovery policy."""
    mock = decoy.mock(cls=ErrorRecoveryPolicy)
    decoy.when(state_store.commands.get_error_recovery_policy()).then_return(mock)
    return mock


def get_next_tracker(
    decoy: Decoy, provider: CommandNoteTrackerProvider
) -> CommandNoteTracker:
    """Get the next tracker provided by a provider, in code without being a fixture.

    This is useful for testing the execution of multiple commands, each of which will get
    a different tracker instance.
    """
    new_tracker = decoy.mock(cls=CommandNoteTracker)
    decoy.when(provider()).then_return(new_tracker)
    return new_tracker


@pytest.fixture
def command_note_tracker(
    decoy: Decoy, command_note_tracker_provider: CommandNoteTrackerProvider
) -> CommandNoteTracker:
    """Get the tracker that the provider will provide."""
    return get_next_tracker(decoy, command_note_tracker_provider)


@pytest.fixture
def subject(
    hardware_api: HardwareControlAPI,
    state_store: StateStore,
    action_dispatcher: ActionDispatcher,
    equipment: EquipmentHandler,
    file_provider: FileProvider,
    camera_provider: CameraProvider,
    movement: MovementHandler,
    mock_gantry_mover: GantryMover,
    labware_movement: LabwareMovementHandler,
    pipetting: PipettingHandler,
    mock_tip_handler: TipHandler,
    run_control: RunControlHandler,
    rail_lights: RailLightsHandler,
    status_bar: StatusBarHandler,
    task_handler: TaskHandler,
    model_utils: ModelUtils,
    command_note_tracker_provider: CommandNoteTrackerProvider,
) -> CommandExecutor:
    """Get a CommandExecutor test subject with its dependencies mocked out."""
    return CommandExecutor(
        hardware_api=hardware_api,
        file_provider=file_provider,
        camera_provider=camera_provider,
        state_store=state_store,
        action_dispatcher=action_dispatcher,
        equipment=equipment,
        movement=movement,
        gantry_mover=mock_gantry_mover,
        labware_movement=labware_movement,
        pipetting=pipetting,
        tip_handler=mock_tip_handler,
        run_control=run_control,
        model_utils=model_utils,
        rail_lights=rail_lights,
        status_bar=status_bar,
        task_handler=task_handler,
        command_note_tracker_provider=command_note_tracker_provider,
    )


class _TestCommandParams(BaseModel):
    foo: str = "foo"


class _TestCommandResult(BaseModel):
    bar: str = "bar"


class _TestCommandDefinedError(ErrorOccurrence):
    errorType: str = "testCommandDefinedError"
    detail: str = "test command defined error"


_TestCommandReturn = Union[
    SuccessData[_TestCommandResult],
    DefinedErrorData[_TestCommandDefinedError],
]


class _TestCommandImpl(AbstractCommandImpl[_TestCommandParams, _TestCommandReturn]):
    async def execute(self, params: _TestCommandParams) -> _TestCommandReturn:
        raise NotImplementedError()


async def test_execute(
    decoy: Decoy,
    hardware_api: HardwareControlAPI,
    state_store: StateStore,
    action_dispatcher: ActionDispatcher,
    equipment: EquipmentHandler,
    file_provider: FileProvider,
    camera_provider: CameraProvider,
    movement: MovementHandler,
    mock_gantry_mover: GantryMover,
    labware_movement: LabwareMovementHandler,
    pipetting: PipettingHandler,
    mock_tip_handler: TipHandler,
    run_control: RunControlHandler,
    rail_lights: RailLightsHandler,
    status_bar: StatusBarHandler,
    task_handler: TaskHandler,
    model_utils: ModelUtils,
    command_note_tracker: CommandNoteTracker,
    subject: CommandExecutor,
) -> None:
    """It should be able to execute a command."""
    TestCommandImplCls = decoy.mock(func=_TestCommandImpl)
    command_impl = decoy.mock(cls=_TestCommandImpl)

    # Note: private attrs (which are attrs that start with _) are instantiated via deep
    # copy from a provided default in the model, so if
    # _TestCommand()._ImplementationCls != _TestCommand._ImplementationCls.default if
    # we provide a default. Therefore, provide a default factory, so we can always have
    # the same object.

    class _TestCommand(
        BaseCommand[_TestCommandParams, _TestCommandResult, ErrorOccurrence]
    ):
        commandType: str = "testCommand"
        params: _TestCommandParams
        result: Optional[_TestCommandResult]

        _ImplementationCls: Type[_TestCommandImpl] = PrivateAttr(
            default_factory=lambda: TestCommandImplCls
        )

    command_params = _TestCommandParams()
    command_result = SuccessData(public=_TestCommandResult())

    queued_command = cast(
        Command,
        _TestCommand.model_construct(  # type: ignore[call-arg]
            id="command-id",
            key="command-key",
            createdAt=datetime(year=2021, month=1, day=1),
            status=CommandStatus.QUEUED,
            params=command_params,
        ),
    )

    running_command = cast(
        Command,
        _TestCommand.model_construct(  # type: ignore[call-arg]
            id="command-id",
            key="command-key",
            createdAt=datetime(year=2021, month=1, day=1),
            startedAt=datetime(year=2022, month=2, day=2),
            status=CommandStatus.RUNNING,
            params=command_params,
        ),
    )

    command_notes = [
        CommandNote(
            noteKind="warning",
            shortMessage="hello",
            longMessage="test command note",
            source="test",
        )
    ]

    expected_completed_command = cast(
        Command,
        _TestCommand.model_construct(
            id="command-id",
            key="command-key",
            createdAt=datetime(year=2021, month=1, day=1),
            startedAt=datetime(year=2022, month=2, day=2),
            completedAt=datetime(year=2023, month=3, day=3),
            status=CommandStatus.SUCCEEDED,
            params=command_params,
            result=command_result.public,
            notes=command_notes,
        ),
    )

    decoy.when(state_store.commands.get(command_id="command-id")).then_return(
        queued_command
    )

    decoy.when(
        action_dispatcher.dispatch(  # type: ignore[func-returns-value]
            RunCommandAction(
                command_id="command-id", started_at=datetime(year=2022, month=2, day=2)
            )
        )
    ).then_do(
        lambda _: decoy.when(
            state_store.commands.get(command_id="command-id")
        ).then_return(running_command)
    )
    decoy.when(
        queued_command._ImplementationCls(
            state_view=state_store,
            hardware_api=hardware_api,
            file_provider=file_provider,
            camera_provider=camera_provider,
            equipment=equipment,
            movement=movement,
            gantry_mover=mock_gantry_mover,
            labware_movement=labware_movement,
            pipetting=pipetting,
            tip_handler=mock_tip_handler,
            run_control=run_control,
            rail_lights=rail_lights,
            model_utils=model_utils,
            status_bar=status_bar,
            task_handler=task_handler,
            command_note_adder=command_note_tracker,
        )
    ).then_return(
        command_impl  # type: ignore[arg-type]
    )

    decoy.when(command_note_tracker.get_notes()).then_return(command_notes)

    decoy.when(await command_impl.execute(command_params)).then_return(command_result)

    decoy.when(model_utils.get_timestamp()).then_return(
        datetime(year=2022, month=2, day=2),
        datetime(year=2023, month=3, day=3),
    )

    decoy.when(state_store.commands.get_error_recovery_policy()).then_return(
        never_recover
    )

    await subject.execute("command-id")

    decoy.verify(
        action_dispatcher.dispatch(
            SucceedCommandAction(command=expected_completed_command)
        ),
    )


@pytest.mark.parametrize(
    ["command_error", "expected_error", "use_camera"],
    [
        (
            errors.ProtocolEngineError(message="oh no"),
            matchers.ErrorMatching(errors.ProtocolEngineError, match="oh no"),
            False,
        ),
        (
            EStopActivatedError(),
            matchers.ErrorMatching(PE_EStopActivatedError),
            False,
        ),
        (
            RuntimeError("oh no"),
            matchers.ErrorMatching(PythonException, match="oh no"),
            False,
        ),
        (
            asyncio.CancelledError(),
            matchers.ErrorMatching(errors.RunStoppedError),
            False,
        ),
        (
            errors.ProtocolEngineError(message="oh no"),
            matchers.ErrorMatching(errors.ProtocolEngineError, match="oh no"),
            True,
        ),
        (
            RuntimeError("oh no"),
            matchers.ErrorMatching(PythonException, match="oh no"),
            True,
        ),
    ],
)
async def test_execute_undefined_error(
    decoy: Decoy,
    hardware_api: HardwareControlAPI,
    state_store: StateStore,
    action_dispatcher: ActionDispatcher,
    equipment: EquipmentHandler,
    file_provider: FileProvider,
    camera_provider: CameraProvider,
    movement: MovementHandler,
    mock_gantry_mover: GantryMover,
    labware_movement: LabwareMovementHandler,
    pipetting: PipettingHandler,
    mock_tip_handler: TipHandler,
    run_control: RunControlHandler,
    rail_lights: RailLightsHandler,
    status_bar: StatusBarHandler,
    task_handler: TaskHandler,
    model_utils: ModelUtils,
    subject: CommandExecutor,
    command_note_tracker: CommandNoteTracker,
    error_recovery_policy: ErrorRecoveryPolicy,
    command_error: Exception,
    expected_error: Any,
    use_camera: bool,
) -> None:
    """It should handle an undefined error raised from execution."""
    TestCommandImplCls = decoy.mock(func=_TestCommandImpl)
    command_impl = decoy.mock(cls=_TestCommandImpl)

    class _TestCommand(
        BaseCommand[_TestCommandParams, _TestCommandResult, ErrorOccurrence]
    ):
        commandType: str = "testCommand"
        params: _TestCommandParams
        result: Optional[_TestCommandResult]

        _ImplementationCls: Type[_TestCommandImpl] = PrivateAttr(
            default_factory=lambda: TestCommandImplCls
        )

    command_params = _TestCommandParams()

    queued_command = cast(
        Command,
        _TestCommand.model_construct(  # type: ignore[call-arg]
            id="command-id",
            key="command-key",
            createdAt=datetime(year=2021, month=1, day=1),
            status=CommandStatus.QUEUED,
            params=command_params,
        ),
    )

    running_command = cast(
        Command,
        _TestCommand.model_construct(  # type: ignore[call-arg]
            id="command-id",
            key="command-key",
            createdAt=datetime(year=2021, month=1, day=1),
            startedAt=datetime(year=2022, month=2, day=2),
            status=CommandStatus.RUNNING,
            params=command_params,
        ),
    )

    command_notes = [
        CommandNote(
            noteKind="warning",
            shortMessage="hello",
            longMessage="test command note",
            source="test",
        )
    ]

    decoy.when(state_store.commands.get(command_id="command-id")).then_return(
        queued_command
    )

    decoy.when(
        action_dispatcher.dispatch(  # type: ignore[func-returns-value]
            RunCommandAction(
                command_id="command-id", started_at=datetime(year=2022, month=2, day=2)
            )
        )
    ).then_do(
        lambda _: decoy.when(
            state_store.commands.get(command_id="command-id")
        ).then_return(running_command)
    )

    decoy.when(
        queued_command._ImplementationCls(
            state_view=state_store,
            hardware_api=hardware_api,
            file_provider=file_provider,
            camera_provider=camera_provider,
            equipment=equipment,
            movement=movement,
            gantry_mover=mock_gantry_mover,
            labware_movement=labware_movement,
            pipetting=pipetting,
            tip_handler=mock_tip_handler,
            run_control=run_control,
            rail_lights=rail_lights,
            model_utils=model_utils,
            status_bar=status_bar,
            task_handler=task_handler,
            command_note_adder=command_note_tracker,
        )
    ).then_return(
        command_impl  # type: ignore[arg-type]
    )

    decoy.when(await command_impl.execute(command_params)).then_raise(command_error)

    decoy.when(model_utils.generate_id()).then_return("error-id")
    decoy.when(model_utils.get_timestamp()).then_return(
        datetime(year=2022, month=2, day=2),
        datetime(year=2023, month=3, day=3),
    )

    decoy.when(
        error_recovery_policy(matchers.Anything(), matchers.Anything(), None)
    ).then_return(ErrorRecoveryType.WAIT_FOR_RECOVERY)

    decoy.when(command_note_tracker.get_notes()).then_return(command_notes)

    if use_camera:
        decoy.when(await subject._camera_provider.get_camera_settings()).then_return(
            CameraSettings(
                cameraEnabled=True,
                liveStreamEnabled=True,
                errorRecoveryCameraEnabled=True,
            )
        )

    await subject.execute("command-id")

    decoy.verify(
        action_dispatcher.dispatch(
            FailCommandAction(
                command_id="command-id",
                running_command=running_command,
                error_id="error-id",
                failed_at=datetime(year=2023, month=3, day=3),
                error=expected_error,
                type=ErrorRecoveryType.WAIT_FOR_RECOVERY,
                notes=command_notes,
            )
        ),
    )

    if use_camera:
        decoy.verify(
            await subject._camera_provider.capture_image(
                subject._state_store.config.robot_type, ImageParameters()
            ),
            times=1,
        )
    else:
        decoy.verify(
            await subject._camera_provider.capture_image(
                subject._state_store.config.robot_type, ImageParameters()
            ),
            times=0,
        )


@pytest.mark.parametrize(
    "use_camera",
    [
        True,
        False,
    ],
)
async def test_execute_defined_error(
    decoy: Decoy,
    subject: CommandExecutor,
    hardware_api: HardwareControlAPI,
    state_store: StateStore,
    action_dispatcher: ActionDispatcher,
    equipment: EquipmentHandler,
    file_provider: FileProvider,
    camera_provider: CameraProvider,
    movement: MovementHandler,
    mock_gantry_mover: GantryMover,
    labware_movement: LabwareMovementHandler,
    pipetting: PipettingHandler,
    mock_tip_handler: TipHandler,
    run_control: RunControlHandler,
    rail_lights: RailLightsHandler,
    status_bar: StatusBarHandler,
    task_handler: TaskHandler,
    model_utils: ModelUtils,
    command_note_tracker: CommandNoteTracker,
    error_recovery_policy: ErrorRecoveryPolicy,
    use_camera: bool,
) -> None:
    """It should handle a defined error returned from execution."""
    TestCommandImplCls = decoy.mock(func=_TestCommandImpl)
    command_impl = decoy.mock(cls=_TestCommandImpl)

    class _TestCommand(
        BaseCommand[_TestCommandParams, _TestCommandResult, ErrorOccurrence]
    ):
        commandType: str = "testCommand"
        params: _TestCommandParams
        result: Optional[_TestCommandResult]

        _ImplementationCls: Type[_TestCommandImpl] = PrivateAttr(
            default_factory=lambda: TestCommandImplCls
        )

    command_params = _TestCommandParams()
    command_id = "command-id"
    created_at = datetime(year=2021, month=1, day=1)
    started_at = datetime(year=2022, month=2, day=2)
    failed_at = datetime(year=2023, month=3, day=3)
    error_id = "error-id"
    returned_error = DefinedErrorData(
        public=_TestCommandDefinedError(id=error_id, createdAt=failed_at),
    )
    queued_command = cast(
        Command,
        _TestCommand.model_construct(  # type: ignore[call-arg]
            id=command_id,
            key="command-key",
            createdAt=created_at,
            status=CommandStatus.QUEUED,
            params=command_params,
        ),
    )
    running_command = cast(
        Command,
        _TestCommand.model_construct(  # type: ignore[call-arg]
            id=command_id,
            key="command-key",
            createdAt=created_at,
            startedAt=started_at,
            status=CommandStatus.RUNNING,
            params=command_params,
        ),
    )
    command_notes = [
        CommandNote(
            noteKind="warning",
            shortMessage="hello",
            longMessage="test command note",
            source="test",
        )
    ]

    decoy.when(state_store.commands.get(command_id=command_id)).then_return(
        queued_command
    )

    decoy.when(
        action_dispatcher.dispatch(  # type: ignore[func-returns-value]
            RunCommandAction(command_id=command_id, started_at=started_at)
        )
    ).then_do(
        lambda _: decoy.when(
            state_store.commands.get(command_id=command_id)
        ).then_return(running_command)
    )

    decoy.when(
        queued_command._ImplementationCls(
            state_view=state_store,
            hardware_api=hardware_api,
            file_provider=file_provider,
            camera_provider=camera_provider,
            equipment=equipment,
            movement=movement,
            gantry_mover=mock_gantry_mover,
            labware_movement=labware_movement,
            pipetting=pipetting,
            tip_handler=mock_tip_handler,
            task_handler=task_handler,
            run_control=run_control,
            rail_lights=rail_lights,
            model_utils=model_utils,
            status_bar=status_bar,
            command_note_adder=command_note_tracker,
        )
    ).then_return(
        command_impl  # type: ignore[arg-type]
    )

    decoy.when(command_note_tracker.get_notes()).then_return(command_notes)

    decoy.when(await command_impl.execute(command_params)).then_return(returned_error)

    decoy.when(model_utils.generate_id()).then_return(error_id)
    decoy.when(model_utils.get_timestamp()).then_return(started_at, failed_at)

    decoy.when(
        error_recovery_policy(
            matchers.Anything(),
            matchers.Anything(),
            returned_error,  # type: ignore[arg-type]
        )
    ).then_return(ErrorRecoveryType.WAIT_FOR_RECOVERY)

    if use_camera:
        decoy.when(await subject._camera_provider.get_camera_settings()).then_return(
            CameraSettings(
                cameraEnabled=True,
                liveStreamEnabled=True,
                errorRecoveryCameraEnabled=True,
            )
        )

    await subject.execute("command-id")

    decoy.verify(
        action_dispatcher.dispatch(
            FailCommandAction(
                command_id="command-id",
                running_command=running_command,
                error_id="error-id",
                failed_at=datetime(year=2023, month=3, day=3),
                error=returned_error,  # type: ignore[arg-type]
                type=ErrorRecoveryType.WAIT_FOR_RECOVERY,
                notes=command_notes,
            )
        )
    )

    if use_camera:
        decoy.verify(
            await subject._camera_provider.capture_image(
                subject._state_store.config.robot_type, ImageParameters()
            ),
            times=1,
        )
    else:
        decoy.verify(
            await subject._camera_provider.capture_image(
                subject._state_store.config.robot_type, ImageParameters()
            ),
            times=0,
        )



import pytest

from opentrons.drivers import utils
from opentrons.drivers.smoothie_drivers import parse_utils
from opentrons.drivers.utils import ParseError


def test_parse_position_response() -> None:
    good_data = "ok M114.2 X:10 Y:20 Z:30 A:40 B:50 C:60"
    res = parse_utils.parse_position_response(good_data)
    expected = {
        "X": 10,
        "Y": 20,
        "Z": 30,
        "A": 40,
        "B": 50,
        "C": 60,
    }
    assert res == expected


@pytest.mark.parametrize(
    argnames=["data"],
    argvalues=[
        [""],
        ["ok M114.2 X:10 Y:20: Z:30A:40 B:50 C:60"],
        ["ok M114.2 X:10 Y:20: Z:30 A:40 B:50"],
    ],
)
def test_parse_position_response_error(data: str) -> None:
    with pytest.raises(ParseError):
        parse_utils.parse_position_response(data)


def test_parse_pipette_data() -> None:
    msg = "TestsRule!!"
    mount = "L"
    good_data = mount + ": " + utils.string_to_hex(msg)
    parsed = parse_utils.parse_instrument_data(good_data).get(mount)
    assert parsed
    assert parsed.decode() == msg


@pytest.mark.parametrize(
    argnames=["data"],
    argvalues=[
        [""],
        ["X:"],
        ["X:4049"],
        ["L:0.212"],
    ],
)
def test_parse_pipette_data_error(data: str) -> None:
    with pytest.raises(ParseError):
        parse_utils.parse_instrument_data(data)



"""Tests for the legacy ModuleGeometry interface."""

from contextlib import nullcontext as does_not_raise
from typing import Any, ContextManager, Optional

import mock
import pytest
from pytest_lazy_fixtures import lf as lazy_fixture

from opentrons_shared_data.module.types import (
    ModuleDefinitionV1,
    ModuleDefinitionV3,
)

from opentrons.hardware_control.modules.types import HeaterShakerModuleModel, ModuleType
from opentrons.protocol_api.core.legacy.deck import Deck
from opentrons.protocol_api.core.legacy.module_geometry import (
    HeaterShakerGeometry,
    ModuleGeometry,
    PipetteMovementRestrictedByHeaterShakerError,
    create_geometry,
)
from opentrons.protocols.api_support.deck_type import STANDARD_OT2_DECK
from opentrons.types import Location, Point


@pytest.fixture
def v1_mag_module_schema_v3_definition() -> ModuleDefinitionV3:
    """A Gen1 Magnetic Module's schemaV3 definition fixture."""
    return {
        "$otSharedSchema": "module/schemas/3",
        "moduleType": "magneticModuleType",
        "model": "magneticModuleV1",
        "labwareOffset": {"x": 11.0, "y": 22.0, "z": 33.0},
        "dimensions": {
            "bareOverallHeight": 123,
            "overLabwareHeight": 234,
            "xDimension": 345,
            "yDimension": 456,
            "labwareInterfaceXDimension": 234,
            "labwareInterfaceYDimension": 123,
        },
        "calibrationPoint": {"x": 111, "y": 222, "z": 333},
        "displayName": "Sample Module",
        "quirks": [],
        "slotTransforms": {},
        "compatibleWith": ["someSimilarModule"],  # type: ignore[list-item]
        "cornerOffsetFromSlot": {"x": 111, "y": 222, "z": 333},
        "config": {},
    }


@pytest.fixture
def minimal_heater_shaker_definition() -> ModuleDefinitionV3:
    """A Heater-shaker module's definition fixture."""
    return {
        "$otSharedSchema": "module/schemas/3",
        "moduleType": "heaterShakerModuleType",
        "model": "heaterShakerModuleV1",
        "labwareOffset": {"x": 11.0, "y": 22.0, "z": 33.0},
        "dimensions": {
            "bareOverallHeight": 123,
            "overLabwareHeight": 234,
            "xDimension": 345,
            "yDimension": 456,
            "labwareInterfaceXDimension": 234,
            "labwareInterfaceYDimension": 234,
        },
        "calibrationPoint": {"x": 111, "y": 222, "z": 333},
        "displayName": "Sample H/S Module",
        "quirks": [],
        "slotTransforms": {},
        "compatibleWith": ["someSimilarModule"],  # type: ignore[list-item]
        "cornerOffsetFromSlot": {"x": 111, "y": 222, "z": 333},
        "config": {},
    }


@pytest.fixture
def v1_mag_module_schema_v1_definition() -> ModuleDefinitionV1:
    """A Gen1 Magnetic Module's schemaV1 definition fixture."""
    return {
        "labwareOffset": {"x": 11.0, "y": 22.0, "z": 33.0},
        "dimensions": {
            "bareOverallHeight": 123,
            "overLabwareHeight": 321,
            "labwareInterfaceXDimension": 12,
            "labwareInterfaceYDimension": 23,
        },
        "calibrationPoint": {"x": 44.0, "y": 55.0},
        "displayName": "Sample Old Module",
        "loadName": "magdeck",
        "quirks": [],
    }


@pytest.fixture
def heater_shaker_geometry() -> HeaterShakerGeometry:
    """Get a Heater-Shaker Geometry fixture."""
    heater_shaker_slot_location = Deck(deck_type=STANDARD_OT2_DECK).position_for(5)
    return HeaterShakerGeometry(
        display_name="A new shiny module!",
        model=HeaterShakerModuleModel.HEATER_SHAKER_V1,
        module_type=ModuleType.HEATER_SHAKER,
        offset=Point(0, 0, 0),
        overall_height=111,
        height_over_labware=222,
        parent=heater_shaker_slot_location,
    )


@pytest.fixture
def mock_location() -> mock.MagicMock:
    """Get a mocked out Location object."""
    return mock.MagicMock(return_value=Location(point=Point(1, 2, 3), labware=None))


@pytest.mark.parametrize(
    argnames=[
        "module_definition",
        "expected_geometry",
        "expected_parent_location",
        "expected_repr",
    ],
    argvalues=[
        (
            lazy_fixture("minimal_heater_shaker_definition"),
            HeaterShakerGeometry(
                parent=Location(Point(0, 0, 0), labware="5"),
                offset=Point(11, 22, 33),
                overall_height=123,
                height_over_labware=234,
                model=HeaterShakerModuleModel.HEATER_SHAKER_V1,
                module_type=ModuleType.HEATER_SHAKER,
                display_name="Sample H/S Module",
            ),
            "5",
            "Sample H/S Module on 5",
        ),
        (
            lazy_fixture("minimal_heater_shaker_definition"),
            HeaterShakerGeometry(
                parent=Location(Point(0, 0, 0), labware=None),
                offset=Point(11, 22, 33),
                overall_height=123,
                height_over_labware=234,
                model=HeaterShakerModuleModel.HEATER_SHAKER_V1,
                module_type=ModuleType.HEATER_SHAKER,
                display_name="Sample H/S Module",
            ),
            None,
            "Sample H/S Module",
        ),
    ],
)
def test_create_geometry(
    module_definition: ModuleDefinitionV3,
    expected_geometry: ModuleGeometry,
    expected_parent_location: Optional[str],
    expected_repr: str,
) -> None:
    """It should load an API-version-specific module from its definition."""
    load_result = create_geometry(
        definition=module_definition,
        parent=Location(point=Point(0, 0, 0), labware=expected_parent_location),
        configuration=None,
    )
    assert isinstance(load_result, expected_geometry.__class__)
    assert load_result.parent == expected_geometry.parent
    assert load_result.module_type == expected_geometry.module_type
    assert load_result.model == expected_geometry.model
    assert load_result.labware_offset == expected_geometry.labware_offset
    assert load_result.highest_z == expected_geometry.highest_z
    assert str(load_result) == expected_repr


def test_create_geometry_raises(
    v1_mag_module_schema_v3_definition: ModuleDefinitionV3,
) -> None:
    """It raises when an invalid definition is passed."""
    v1_mag_module_schema_v3_definition.update({"moduleType": "blahblahModuleType"})  # type: ignore[typeddict-item]

    with pytest.raises(ValueError):
        create_geometry(
            definition=v1_mag_module_schema_v3_definition,
            parent=Location(point=Point(0, 0, 0), labware=None),
            configuration=None,
        )


def test_heater_shaker_geometry_properties() -> None:
    """It should initialize a heater-shaker geometry instance with all properties."""
    subject = HeaterShakerGeometry(
        display_name="A new shiny module!",
        model=HeaterShakerModuleModel.HEATER_SHAKER_V1,
        module_type=ModuleType.HEATER_SHAKER,
        offset=Point(0, 0, 0),
        overall_height=111,
        height_over_labware=222,
        parent=Location(point=Point(1, 2, 3), labware=None),
    )
    assert subject.model == HeaterShakerModuleModel.HEATER_SHAKER_V1
    assert subject.labware is None
    assert subject.location.point == Point(1, 2, 3)


@pytest.mark.parametrize(
    argnames=["destination_slot", "expected_raise"],
    argvalues=[
        [
            4,
            pytest.raises(
                PipetteMovementRestrictedByHeaterShakerError, match="shaking"
            ),
        ],  # east
        [
            6,
            pytest.raises(
                PipetteMovementRestrictedByHeaterShakerError, match="shaking"
            ),
        ],  # west
        [
            8,
            pytest.raises(
                PipetteMovementRestrictedByHeaterShakerError, match="shaking"
            ),
        ],  # north
        [
            2,
            pytest.raises(
                PipetteMovementRestrictedByHeaterShakerError, match="shaking"
            ),
        ],  # south
        [
            5,
            pytest.raises(
                PipetteMovementRestrictedByHeaterShakerError, match="shaking"
            ),
        ],  # h/s
        [1, does_not_raise()],  # non-adjacent
    ],
)
def test_hs_raises_when_moving_to_restricted_slots_while_shaking(
    heater_shaker_geometry: HeaterShakerGeometry,
    destination_slot: int,
    expected_raise: ContextManager[Any],
) -> None:
    """It should raise if restricted movement around a heater-shaker is attempted while module is shaking."""
    with expected_raise:
        heater_shaker_geometry.flag_unsafe_move(
            to_slot=destination_slot,
            is_tiprack=False,
            is_using_multichannel=False,
            is_labware_latch_closed=True,
            is_plate_shaking=True,
        )


@pytest.mark.parametrize(
    argnames=["destination_slot", "expected_raise"],
    argvalues=[
        [
            4,
            pytest.raises(PipetteMovementRestrictedByHeaterShakerError, match="latch"),
        ],  # east
        [
            6,
            pytest.raises(PipetteMovementRestrictedByHeaterShakerError, match="latch"),
        ],  # west
        [
            5,
            pytest.raises(PipetteMovementRestrictedByHeaterShakerError, match="latch"),
        ],  # h/s
        [8, does_not_raise()],  # north
        [2, does_not_raise()],  # south
        [3, does_not_raise()],  # non-adjacent
    ],
)
def test_raises_when_moving_to_restricted_slots_while_latch_open(
    heater_shaker_geometry: HeaterShakerGeometry,
    destination_slot: int,
    expected_raise: ContextManager[Any],
) -> None:
    """It should raise if restricted movement around a heater-shaker is attempted while latch is open."""
    with expected_raise:
        heater_shaker_geometry.flag_unsafe_move(
            to_slot=destination_slot,
            is_tiprack=False,
            is_using_multichannel=False,
            is_labware_latch_closed=False,
            is_plate_shaking=False,
        )


@pytest.mark.parametrize(
    argnames=["destination_slot", "is_tiprack", "expected_raise"],
    argvalues=[
        [
            4,
            False,
            pytest.raises(
                PipetteMovementRestrictedByHeaterShakerError, match="left or right"
            ),
        ],  # east
        [
            6,
            False,
            pytest.raises(
                PipetteMovementRestrictedByHeaterShakerError, match="left or right"
            ),
        ],  # west
        [
            8,
            False,
            pytest.raises(
                PipetteMovementRestrictedByHeaterShakerError,
                match="non-tip-rack labware",
            ),
        ],  # north, non-tiprack
        [
            2,
            False,
            pytest.raises(
                PipetteMovementRestrictedByHeaterShakerError,
                match="non-tip-rack labware",
            ),
        ],  # south, non-tiprack
        [8, True, does_not_raise()],  # north, tiprack
        [2, True, does_not_raise()],  # south, tiprack
        [5, False, does_not_raise()],  # h/s
        [7, False, does_not_raise()],  # non-adjacent
    ],
)
def test_raises_on_restricted_movement_with_multi_channel(
    heater_shaker_geometry: HeaterShakerGeometry,
    destination_slot: int,
    is_tiprack: bool,
    expected_raise: ContextManager[Any],
) -> None:
    """It should raise if restricted movement around a heater-shaker is attempted with a multi-channel pipette."""
    with expected_raise:
        heater_shaker_geometry.flag_unsafe_move(
            to_slot=destination_slot,
            is_tiprack=is_tiprack,
            is_using_multichannel=True,
            is_labware_latch_closed=True,
            is_plate_shaking=False,
        )


@pytest.mark.parametrize(
    argnames=["destination_slot"],
    argvalues=[
        [4],  # east
        [6],  # west
        [5],  # h/s
        [8],  # north
        [2],  # south
        [9],  # non-adjacent
    ],
)
def test_does_not_raise_when_idle_and_latch_closed(
    heater_shaker_geometry: HeaterShakerGeometry,
    destination_slot: int,
) -> None:
    """It should not raise if single channel pipette moves anywhere near heater-shaker when idle and latch closed."""
    with does_not_raise():
        heater_shaker_geometry.flag_unsafe_move(
            to_slot=destination_slot,
            is_tiprack=False,
            is_using_multichannel=False,
            is_labware_latch_closed=True,
            is_plate_shaking=False,
        )


@pytest.mark.parametrize(
    argnames=["pipette_slot", "expected_is_blocking"],
    argvalues=[
        ("4", True),
        ("6", True),
        ("2", True),
        ("5", True),
        ("8", True),
        (None, True),
        ("1", False),
    ],
)
def test_pipette_is_blocking_shake_movement(
    heater_shaker_geometry: HeaterShakerGeometry,
    mock_location: mock.MagicMock,
    pipette_slot: Optional[str],
    expected_is_blocking: bool,
) -> None:
    """It should return True if pipette is blocking shake movement."""
    mock_location.labware.first_parent = mock.MagicMock(return_value=pipette_slot)

    assert (
        heater_shaker_geometry.is_pipette_blocking_shake_movement(
            pipette_location=mock_location
        )
        == expected_is_blocking
    )


@pytest.mark.parametrize(
    argnames=["pipette_slot", "expected_is_blocking"],
    argvalues=[("4", True), ("6", True), ("2", False), (None, True), ("1", False)],
)
def test_pipette_is_blocking_latch_movement(
    heater_shaker_geometry: HeaterShakerGeometry,
    mock_location: mock.MagicMock,
    pipette_slot: Optional[str],
    expected_is_blocking: bool,
) -> None:
    """It should return True if pipette is blocking latch movement."""
    mock_location.labware.first_parent = mock.MagicMock(return_value=pipette_slot)

    assert (
        heater_shaker_geometry.is_pipette_blocking_latch_movement(
            pipette_location=mock_location
        )
        == expected_is_blocking
    )


def test_pipette_is_blocking_shake_and_latch_movements_with_no_pipette_slot(
    heater_shaker_geometry: HeaterShakerGeometry,
    mock_location: mock.MagicMock,
) -> None:
    """It should return True if pipette's last location slot is not known."""
    assert (
        heater_shaker_geometry.is_pipette_blocking_shake_movement(
            pipette_location=Location(point=Point(3, 2, 1), labware=None)
        )
        is True
    )
    assert (
        heater_shaker_geometry.is_pipette_blocking_latch_movement(
            pipette_location=Location(point=Point(3, 2, 1), labware=None)
        )
        is True
    )



"""Smoke tests for the ProtocolEngine creation factory."""

import pytest
from pytest_lazy_fixtures import lf as lazy_fixture

from opentrons_shared_data.deck import load as load_deck
from opentrons_shared_data.deck.types import DeckDefinitionV5
from opentrons_shared_data.labware.labware_definition import LabwareDefinition
from opentrons_shared_data.robot.types import RobotType

from opentrons.calibration_storage.helpers import uri_from_details
from opentrons.hardware_control import API as HardwareAPI
from opentrons.hardware_control.types import DoorState
from opentrons.protocol_engine import (
    Config as EngineConfig,
)
from opentrons.protocol_engine import (
    DeckType,
    ProtocolEngine,
    error_recovery_policy,
)
from opentrons.protocol_engine.create_protocol_engine import create_protocol_engine
from opentrons.protocol_engine.types import DeckSlotLocation, LoadedLabware
from opentrons.protocols.api_support.deck_type import (
    SHORT_TRASH_DECK,
    STANDARD_OT2_DECK,
    STANDARD_OT3_DECK,
)
from opentrons.types import DeckSlotName


@pytest.fixture(scope="session")
def ot2_standard_deck_def() -> DeckDefinitionV5:
    """Get the OT-2 standard deck definition."""
    return load_deck(STANDARD_OT2_DECK, 5)


@pytest.fixture(scope="session")
def ot2_short_trash_deck_def() -> DeckDefinitionV5:
    """Get the OT-2 with short trash standard deck definition."""
    return load_deck(SHORT_TRASH_DECK, 5)


@pytest.fixture(scope="session")
def ot3_standard_deck_def() -> DeckDefinitionV5:
    """Get the OT-2 standard deck definition."""
    return load_deck(STANDARD_OT3_DECK, 5)


@pytest.mark.parametrize(
    (
        "robot_type",
        "deck_type",
        "expected_deck_def",
    ),
    [
        (
            "OT-2 Standard",
            DeckType.OT2_STANDARD,
            lazy_fixture("ot2_standard_deck_def"),
        ),
        (
            "OT-2 Standard",
            DeckType.OT2_SHORT_TRASH,
            lazy_fixture("ot2_short_trash_deck_def"),
        ),
        (
            "OT-3 Standard",
            DeckType.OT3_STANDARD,
            lazy_fixture("ot3_standard_deck_def"),
        ),
    ],
)
async def test_create_engine_initializes_state_with_no_fixed_trash(
    hardware_api: HardwareAPI,
    robot_type: RobotType,
    deck_type: DeckType,
    expected_deck_def: DeckDefinitionV5,
) -> None:
    """It should load deck geometry data into the store on create."""
    engine = await create_protocol_engine(
        hardware_api=hardware_api,
        config=EngineConfig(
            # robot_type chosen to match hardware_api.
            robot_type=robot_type,
            deck_type=deck_type,
        ),
        error_recovery_policy=error_recovery_policy.never_recover,
        load_fixed_trash=False,
    )
    state = engine.state_view

    assert isinstance(engine, ProtocolEngine)
    assert state.labware.get_deck_definition() == expected_deck_def
    assert state.labware.get_all() == []


@pytest.mark.parametrize(
    (
        "robot_type",
        "deck_type",
        "expected_deck_def",
        "expected_fixed_trash_def",
        "expected_fixed_trash_slot",
    ),
    [
        (
            "OT-2 Standard",
            DeckType.OT2_STANDARD,
            lazy_fixture("ot2_standard_deck_def"),
            lazy_fixture("ot2_fixed_trash_def"),
            DeckSlotName.FIXED_TRASH,
        ),
        (
            "OT-2 Standard",
            DeckType.OT2_SHORT_TRASH,
            lazy_fixture("ot2_short_trash_deck_def"),
            lazy_fixture("ot2_short_fixed_trash_def"),
            DeckSlotName.FIXED_TRASH,
        ),
        (
            "OT-3 Standard",
            DeckType.OT3_STANDARD,
            lazy_fixture("ot3_standard_deck_def"),
            lazy_fixture("ot3_fixed_trash_def"),
            DeckSlotName.SLOT_A3,
        ),
    ],
)
async def test_create_engine_initializes_state_with_fixed_trash(
    hardware_api: HardwareAPI,
    robot_type: RobotType,
    deck_type: DeckType,
    expected_deck_def: DeckDefinitionV5,
    expected_fixed_trash_def: LabwareDefinition,
    expected_fixed_trash_slot: DeckSlotName,
) -> None:
    """It should load deck geometry data into the store on create."""
    engine = await create_protocol_engine(
        hardware_api=hardware_api,
        config=EngineConfig(
            # robot_type chosen to match hardware_api.
            robot_type=robot_type,
            deck_type=deck_type,
        ),
        error_recovery_policy=error_recovery_policy.never_recover,
        load_fixed_trash=True,
    )
    state = engine.state_view

    assert isinstance(engine, ProtocolEngine)
    assert state.labware.get_deck_definition() == expected_deck_def
    assert state.labware.get_all() == [
        LoadedLabware(
            id="fixedTrash",
            loadName=expected_fixed_trash_def.parameters.loadName,
            definitionUri=uri_from_details(
                load_name=expected_fixed_trash_def.parameters.loadName,
                namespace=expected_fixed_trash_def.namespace,
                version=expected_fixed_trash_def.version,
            ),
            location=DeckSlotLocation(slotName=expected_fixed_trash_slot),
            offsetId=None,
        )
    ]


async def test_create_engine_initializes_state_with_door_state(
    hardware_api: HardwareAPI,
) -> None:
    """It should load current door status into the store on create."""
    hardware_api.door_state = DoorState.OPEN
    engine = await create_protocol_engine(
        hardware_api=hardware_api,
        config=EngineConfig(
            block_on_door_open=True,
            # Choice of robot and deck type are arbitrary.
            robot_type="OT-2 Standard",
            deck_type=DeckType.OT2_SHORT_TRASH,
        ),
        error_recovery_policy=error_recovery_policy.never_recover,
    )
    state = engine.state_view
    assert state.commands.get_is_door_blocking() is True
